# #!/usr/bin/env python

# """Methods for extracting partitions from trees.

# Formatting
# -----------
# We considered the following formats for partitions/bipartitions:

# ({0, 1}, {2, 3}) - fastest, (node, node.up) sorted, but not easy to compare between trees.
# ((0, 1), (2, 3)) - sorting of biparts and parts allows for easy comparison between trees.
# {(0, 1), (2, 3)} - sorting of biparts destroys order within parts.
# {{0, 1}, {2, 3}} - XXX not possible.

# The first and second formats both have useful cases and so we provide
# functions to return both types.

# Examples
# --------
# Methods for examining splits on a tree:


# Get sets of Nodes on either side of each edge in a tree.
# >>> tree.iter_edge_bipartition_sets()      # ({0, 1}, {2, 3, 4}), ...

# Get sets of Nodes from the four subclades of each edge in a tree.
# >>> tree.iter_edge_quadripartition_sets()  # (({0}, {1}), ({2}, {3, 4})), ...

# Get name-ordered tuples of Nodes on either side of each edge in a tree.
# >>> tree.iter_bipartitions()               # ((0, 1), (2, 3, 4)), ...

# Get name-ordered tuples of Nodes for each quartet induced by bipartitions in a tree.
# >>> tree.iter_quartets()                   # ((0, 1), (2, 3)), ...

# Get name-ordered tuples of Nodes for each quartet induced by quadripartitions in a tree.
# >>> tree.iter_quadripartitions()             # ((0, 1), (2, 3)), ...

# Get fast unordered sets of all combinations of 4 tip Nodes in a tree
# >>> tree._iter_quartet_sets()              # {0, 1, 2, 3}, ...

# Get number of quartets induced by the splits in a tree.
# >>> tree.get_n_quartets()                  # 5

# Get bipartition for a specific node/edge.
# >>> tree.get_bipartition_for_edge(tree, 1)  # ((0, 1), (2, 3, 4)), ...

# Get quartets for a specific node/edge.
# >>> tree.get_quartets_for_edge(tree, 1)      # [((0, 1), (2, 3)), ...]

# Get quadripartitions for a specific node/edge.
# >>> tree.get_quadripartitions_for_edge(tree, 1)  # [((0, 1), (2, 3)), ...]

# TODO
# ----
# after revisiting the use of bipartitions etc in distance methods
# come back to this code and decide if tuples or sets should be used,
# or both as an option.




# DEPRECATED!!!!



# """

# # pylint: disable='too-many-branches'

# from typing import TypeVar, Iterator, Tuple, Optional, Set, List, Union, Callable, Sequence, Any
# import itertools
# from loguru import logger
# import pandas as pd
# import numpy as np
# from toytree import Node, ToyTree
# from toytree.core.apis import TreeEnumAPI, add_subpackage_method, add_toytree_method
# from toytree.utils import ToytreeError

# logger = logger.bind(name="toytree")
# Query = TypeVar("Query")
# # Node = TypeVar("Node")

# __all__ = [
#     "iter_edges",
#     "iter_bipartitions",
#     # "_iter_unordered_quadripartitions",
#     "iter_quadripartitions",
#     "iter_quartets",
#     # "iter_edge_quadripartition_sets",
#     # "iter_bipartitions_info",
#     "get_edges",
#     "get_bipartitions_table",
#     "get_bipartition_for_edge",
# ]





# @add_toytree_method(ToyTree)
# @add_subpackage_method(TreeEnumAPI)
# def iter_edges(self, feature: Optional[str] = None) -> Iterator[Tuple[Node, Node]]:
#     """Generator of (Node, Node) tuples representing edges in tree.

#     Given the current tree rooting edges are yielded in idx order
#     as (child, parent). The edges set will include edges connected
#     to the root Node even if the tree is unrooted. To instead view
#     edges as bipartitions of Node sets see `iter_bipartitions`.

#     Parameters
#     ----------
#     feature: str
#         An optional feature of a Node returned in place of the Node
#         object to represent it. For example, 'name' or 'idx'.

#     Example
#     -------
#     >>> tree = toytree.rtree.unittree(5, seed=123)
#     >>> list(iter_edges(feature='idx'))
#     >>> # [(0, 5), (1, 5), (2, 6), (3, 7), (4, 7), (5, 6), (6, 8), (7, 8)]
#     """
#     if feature:
#         for node in self[:-1]:
#             yield (getattr(node, feature), getattr(node._up, feature))
#     else:
#         for node in self[:-1]:
#             yield (node, node._up)


# @add_toytree_method(ToyTree)
# @add_subpackage_method(TreeEnumAPI)
# def get_edges(self, feature: Optional[str] = None, df: bool = False) -> Union[np.ndarray, pd.DataFrame]:
#     """Return matrix of (child, parent) edges.

#     Given the current tree rooting edges are yielded in idx order
#     as (child, parent).

#     Parameters
#     ----------
#     feature: str or None
#         Edges can be represented by Node features such as 'idx' or
#         'name' (default='idx'). None returns Node objects.
#     df: bool
#         If True the matrix is returned as a pd.DataFrame rather
#         than as a np.ndarray.

#     See Also
#     --------
#     tree.iter_edges
#     """
#     edges = list(self.iter_edges(feature=feature))
#     if df:
#         return pd.DataFrame(edges, columns=["child", "parent"])
#     return np.array(edges)


# @add_subpackage_method(TreeEnumAPI)
# def iter_edge_quadripartition_sets(
#     tree: ToyTree,
#     feature: Optional[str] = "name",
#     contract_partitions: bool = False,
#     include_internal_nodes: bool = False,
# ) -> Iterator[Tuple[Tuple[Set, Set], Tuple[Set, Set]]]:
#     """Yield a tuple of sets of tips for each quadripartition.

#     Edge sets are returned as (({e0},{e1}), ({e2},{e3})) representing
#     sets of (tip) Nodes descending from each of the four edges
#     subtending a focal edge, where the focal edge splits e0,e1 from
#     e2,e3. Note the use of sets to indicate that the items are not
#     sorted. Rooting only affects the order in which quadripartition
#     sets are yielded.

#     Parameters
#     ----------
#     tree: ToyTree
#         A tree to extract edge sets from.
#     feature: str or None
#         Feature to represent Nodes in returned sets. Default is 'name'
#         to use Node name strings. None will return Node objects.
#     contract_partitions: bool
#         If True then each partition is contracted to show only the
#         first Node closest to the edge.
#     include_internal_nodes: bool
#         If True then all nodes in each partition are shown, whether it
#         is internal or a tip Node. If False then only tips are shown.
#         Note: This option is overriden if contract_partitions=True in
#         which case the closest Node is shown whether or not it is a tip.

#     Examples
#     --------
#     >>> newick = "((a,b)X,((c,d)Y,e)Z)R;"
#     >>> tree = toytree.tree(newick)

#     >>> sorted(iter_edge_quadripartition_sets(tree))
#     >>> # (({a},{b}),({c,d},{e}))
#     >>> # (({a,b},{e}),({c},{d}))

#     >>> kwargs = dict(exclude_internal_nodes=False)
#     >>> sorted(iter_edge_quadripartition_sets(tree, **kwargs))
#     >>> # (({a},{b}),({c,d,Y},{e}))
#     >>> # (({a,b,X},{e}),({c},{d}))

#     >>> kwargs = dict(expand_partitions=False)
#     >>> sorted(iter_edge_quadripartition_sets(tree, **kwargs))
#     >>> # (({a},{b}),({Y},{e}))
#     >>> # (({X},{e}),({c},{d}))
#     """
#     # tree = tree.unroot()
#     cache = {}

#     # build a cache of descendants for each Node.
#     topnode = tree.nnodes - 1 if tree.is_rooted() else tree.nnodes
#     for node in tree[:topnode]:  # [:-1]:#.traverse("idxorder"):

#         # if tip simply store its info
#         if node.is_leaf():
#             cache[node] = {node}
#             continue

#         # get tips from each child, and store union to this node
#         vals = [cache[i] for i in node.children]

#         # cache only the closest Node (exclude_internal_nodes ignored)
#         if contract_partitions:
#             cache[node] = {node}

#         # cache tip Nodes with or without internal Nodes
#         else:
#             # cache only the tip Nodes
#             if not include_internal_nodes:
#                 cache[node] = set.union(*vals)
#             # cache all nodes including internals
#             else:
#                 cache[node] = set.union(*vals) | {node}

#     #print('x', cache[tree.get_nodes("X")[0]]) TODO
#     # get set of all relevant nodes for set diffs below
#     if include_internal_nodes:
#         nodes = set(cache)
#     else:
#         nodes = set(tree.get_nodes()[:tree.ntips])

#     # iterate over internal nodes/edges in the tree
#     topnode = tree.nnodes - 2 if tree.is_rooted() else tree.nnodes - 1
#     for node in tree[tree.ntips: topnode]:  # tree.nnodes - 1]:
#         # get >= 2 sets of nodes for children below
#         below = [cache[i] for i in node.children]

#         # if up is root then above_left contains all remaining clades
#         if node._up.is_root():
#             above = []
#             for sis in node.get_sisters():
#                 sis_children = sis._children
#                 if sis_children:
#                     for sis_child in sis_children:
#                         above.append(cache[sis_child])
#                 else:
#                     above.append(cache[sis])
#             # above = [cache[sis_child] for sis in node.get_sisters()]
#             # above = above_left
#             # above = [set(itertools.chain(*above_left))]
#             # above_right = []
#             # TODO: need to expand partitions near root?
#         else:
#             # print(node, below, above_left)
#             # get >= 1 sets of nodes to represent >= 1 clades above.
#             above_left = [cache[sis] for sis in node.get_sisters()]
#             above_right = nodes - {node, node._up} - set.union(*above_left + below)

#             # represent final clade by all its desc, or by its mrca.
#             if contract_partitions:
#                 above = above_left + [{tree.get_mrca_node(*above_right)}]
#             else:
#                 above = above_left + [above_right]

#         print("node", node, below, above, "\n") # above_left, above_right, "\n")
#         # add others
#         bpair = itertools.combinations(below, 2)
#         apair = itertools.combinations(above, 2)
#         for x1, x2 in itertools.product(bpair, apair):
#             # require all to exclude root in unrooted tree: ()
#             if all([x1[0], x1[1], x2[0], x2[1]]):
#                 if feature is None:
#                     yield ((x1[0], x1[1]), (x2[0], x2[1]))
#                 else:
#                     yield (
#                         tuple({getattr(j, feature) for j in i} for i in x1),
#                         tuple({getattr(j, feature) for j in i} for i in x2),
#                     )




# @add_subpackage_method(TreeEnumAPI)
# def iter_quadripartitions(
#     tree: ToyTree,
#     feature: Optional[str] = "name",
#     collapse: bool = False,
# ) -> Iterator:
#     r"""Generator yields name-sorted quadripartitions induced by each edge.

#     Quadripartitions are samples of 4 tips in a tree that are drawn
#     from the four subclades induced by an edge in a tree. This
#     collection of splits is a subset of the number of *quartets* in a
#     tree, which can be generated using the `iter_quartets` function.
#     Whereas quartets represent combinations of two samples from either
#     side of bipartition, this function returns combinations of samples
#     drawn from the four clades of a quadripartition induced by a split
#     in the tree. The difference between these two collections is easy
#     to see on larger trees where this set is much smaller than the
#     number of quartets.

#     Example
#     -------
#     >>> tree = toytree.tree("((a,b)X,((c,d)Y,e)Z)R;")
#     >>> sorted(toytree.enumerate.iter_quadripartitions(tree))
#     >>> [(('a', 'b'), ('c', 'e')),
#     >>>  (('a', 'b'), ('d', 'e')),
#     >>>  (('a', 'e'), ('c', 'd')),
#     >>>  (('b', 'e'), ('c', 'd'))]

#     See Also
#     --------
#     - `toytree.enumerate.iter_quartets`
#     - `ToyTree.iter_quartets`
#     """
#     for split in iter_edge_quadripartition_sets(tree, feature=None):
#         for tips in itertools.product(*split[0] + split[1]):
#             tup0 = sorted((tips[0], tips[1]), key=lambda x: x.name)
#             tup1 = sorted((tips[2], tips[3]), key=lambda x: x.name)
#             tups = sorted((tup0, tup1), key=lambda x: x[0].name)
#             qrt = tuple(tups[0] + tups[1])

#             if feature:
#                 qrt = tuple(getattr(i, feature) for i in qrt)

#             if collapse:
#                 yield (qrt[0], qrt[1], qrt[2], qrt[3])
#             else:
#                 yield (qrt[0], qrt[1]), (qrt[2], qrt[3])


# def iter_bipartitions_info(
#     tree: ToyTree,
#     feature: str = "name",
#     include_internal_nodes: bool = False,
#     include_singleton_partitions: bool = False,
#     include_root: bool = False,
# ) -> Iterator[Tuple[Node, Tuple, float]]:
#     """..."""
#     is_rooted = tree.is_rooted()
#     iter_biparts = tree.iter_bipartitions(
#         include_internal_nodes=include_internal_nodes,
#         include_singleton_partitions=True,
#         include_root=include_root,
#     )
#     for node, bipart in zip(tree, iter_biparts):
#         if node._is_leaf():
#             if include_singleton_partitions:
#                 yield node, bipart, node._dist
#         elif node._up.is_root() and is_rooted:
#             yield node, bipart, sum(i._dist for i in node._up.children)
#         else:
#             yield node, bipart, node._dist


# @add_subpackage_method(TreeEnumAPI)
# def get_bipartitions_table(
#     tree: ToyTree,
#     feature: str = "name",
#     include_internal_nodes: bool = False,
#     include_singleton_partitions: bool = False,
# ) -> pd.DataFrame:
#     """Return a DataFrame with partitions from tree in idx order.

#     Partitions represent splits that separate sets of Nodes in a
#     tree, and can be represented by the tips descended from each
#     side of the split, e.g., [['a', 'b'], ['c', 'd']]. Options are
#     available to return all Nodes on either side of a partition,
#     instead of just the tips, but tips are generally of interest.
#     The index of the returned DataFrame corresponds to the Node
#     idx label below the edge of each partition.

#     Note
#     ----
#     The root Node is ignored, and so rooting has no effect on
#     the partitions. This is because the root separates None from
#     all, e.g., [[], ['a', 'b', 'c', 'd']], and the nodes on
#     either side of the root have the same partition, [['a', 'b'],
#     ['c', 'd']], only one of which is returned.

#     Parameters
#     ----------
#     feature: str
#         The Node feature to return for every Node on each side of
#         a split. Default is "name".
#     exclude_internal_labels: bool
#         Default is to only show tip Nodes on either side of a
#         bipartition, but internal Nodes can be included as well.
#     exclude_singleton_splits: bool
#         Default is to exclude singleton splits (e.g., {A | B,C,D})
#         since it is implicit that one exists for every tip Node,
#         but these can be included if requested.

#     See Also
#     --------
#     `ToyTree.iter_bipartitions`, `ToyTree._get_bipartitions_table`

#     Examples
#     --------
#     >>> tree = toytree.rtree.unittree(4)
#     >>> tree.get_bipartitions_table()
#     >>> #     0            1      dist
#     >>> #   (,0)   (1, 2, 3)       0.1
#     >>> #   (,1)      (2, 3)       0.5
#     >>> #   ...
#     """
#     biparts = list(
#         tree.iter_bipartitions(
#             feature=feature,
#             include_internal_nodes=include_internal_nodes,
#             include_singleton_partitions=include_singleton_partitions,
#         )
#     )
#     if not include_singleton_partitions:
#         index = range(tree.ntips, tree.ntips + len(biparts))
#     else:
#         index = None
#     frame = pd.DataFrame(biparts, index=index)

#     # add dist info as unrooted tree
#     frame['dist'] = 0.
#     is_rooted = tree.is_rooted()
#     for idx in frame.index:
#         node = tree[idx]
#         if node._up.is_root() and is_rooted:
#             frame.loc[idx, "dist"] = sum(i._dist for i in node._up.children)
#         else:
#             frame.loc[idx, "dist"] = node._dist
#     return frame


# @add_subpackage_method(TreeEnumAPI)
# def get_bipartition_for_edge(
#     tree: ToyTree,
#     *edge: Query,
#     feature: Optional[str] = "name",
#     include_internal_nodes: bool = False,
# ) -> Tuple[Tuple, Tuple]:
#     """Return the bipartition for a selected Node/edge.

#     Parameters
#     ----------
#     tree: ToyTree
#         A ToyTree from which the bipartition will be found.
#     edge: Node
#         Select an edge on the tree by entering one or more Node queries
#         (Nodes, int Node idx labels, or str Node names) from which the
#         mrca Node will be selected. The edge connects Node and Node.up.
#     feature: str or None
#         Feature used to represent Nodes in the returned bipartition. If
#         None then Node objects are returned. Default is 'name'.
#     include_internal_nodes: bool
#         If True then all Nodes in the biparition are included in the
#         results, else only tip Nodes are included.

#     Examples
#     --------
#     >>> tree = toytree.tree("(a,b,((c,d)Y,(e,f)X)Z)R;")
#     >>> get_bipartition_for_edge(tree, "Y")
#     >>> # (('c', 'd'), ('a', 'b', 'e', 'f'))
#     >>> get_bipartition_for_edge(tree, "Y", exclude_internal_nodes=False)
#     >>> # (('c', 'd', 'Y'), ('a', 'b', 'e', 'f', 'X', 'Z', 'R'))
#     >>> get_bipartition_for_edge(tree, 'a', 'b', exclude_internal_nodes=False)
#     """
#     # get Node to represent selected edge on tree, if root was selected
#     # then just sample its first child since its the same split.
#     edge = tree.get_mrca_node(*edge)
#     if edge.is_root():
#         raise ToytreeError("Cannot select the root, it is not a resolved edge.")

#     # NOTE: SIMPLER CODE POSSIBLE BUT WAY SLOWER B/C ITERATION OVER ALL NODES
#     # for idx, bipart in enumerate(iter_bipartitions(**kwargs)):
#     #     if idx == edge.idx:
#     #         return bipart

#     # a format funtion to return a tuple of Nodes in the requested feature
#     if feature is not None:
#         def tformat(node_tuple):
#             return tuple(getattr(i, feature) for i in node_tuple)
#     else:
#         tformat = tuple

#     # build the bipart for selected edge.
#     if include_internal_nodes:
#         left = set(tree[edge.idx].get_descendants())
#         right = set(tree) - left - {tree.treenode}
#     else:
#         left = set(tree[edge.idx].get_leaves())
#         right = set(tree[:tree.ntips]) - left

#     # sort Node by names within each
#     left = sorted(left, key=_get_node_name_for_sorter)
#     right = sorted(right, key=_get_node_name_for_sorter)

#     # sort biparts to (shorter, longer), or by name if same len.
#     lenl = len(left)
#     lenr = len(right)
#     if lenl < lenr:
#         clade1, clade2 = tformat(left), tformat(right)
#     elif lenr < lenl:
#         clade1, clade2 = tformat(right), tformat(left)
#     else:
#         clade1, clade2 = sorted((left, right), key=lambda x: x[0].name)
#         clade1, clade2 = tformat(clade1), tformat(clade2)
#     return clade1, clade2


# @add_subpackage_method(TreeEnumAPI)
# def get_quartets_for_edge(
#     tree: ToyTree,
#     *edge: Query,
#     feature: Optional[str] = "name",
#     collapse: bool = False,
# ) -> List[Tuple[Tuple, Tuple]]:
#     """..."""
#     quartets = set()
#     bipart = get_bipartition_for_edge(tree, *edge, feature=None)
#     pairgen = itertools.product(
#         itertools.combinations(bipart[0], 2),
#         itertools.combinations(bipart[1], 2),
#     )
#     for tups in pairgen:
#         tup0 = sorted(tups[0], key=lambda x: x.name)
#         tup1 = sorted(tups[1], key=lambda x: x.name)
#         tups = sorted([tup0, tup1], key=lambda x: x[0].name)
#         qrt = tuple(tups[0] + tups[1])
#         # convert feature
#         if feature:
#             qrt = tuple(getattr(i, feature) for i in qrt)
#         # format func: return quartet as ((,),(,)) or (,,,).
#         if not collapse:
#             qrt = tuple((qrt[:2], qrt[2:]))
#         quartets.add(qrt)
#     # if not quartets:
#     #     logger.warning(f"No quartets exist for edge: {edge}")
#     return sorted(quartets)


# @add_subpackage_method(TreeEnumAPI)
# def get_quadripartitions_for_edge(
#     tree: ToyTree,
#     *edge: Query,
#     feature: Optional[str] = "name",
#     contract_partitions: bool = False,
#     include_internal_nodes: bool = False,
# ) -> List[Tuple[Tuple, Tuple]]:
#     """Return the bipartition for a selected Node/edge.

#     Parameters
#     ----------
#     tree: ToyTree
#         A ToyTree from which the bipartition will be found.
#     edge: Node
#         Select an edge on the tree by entering one or more Node queries
#         (Nodes, int Node idx labels, or str Node names) from which the
#         mrca Node will be selected. The edge connects Node and Node.up.
#     feature: str or None
#         Feature used to represent Nodes in the returned bipartition. If
#         None then Node objects are returned. Default is 'name'.

#     Examples
#     --------
#     >>> tree = toytree.tree("(a,b,((c,d)CD,(e,f)EF)X)AB;")
#     >>> sorted(get_quadripartitions_for_edge(tree, "X"))
#     >>> # [(...),
#     >>> #  (...),
#     >>> #  (...)]
#     """
#     # only internal edges form valid quadripartitions.
#     edge = tree.get_mrca_node(*edge)
#     if edge not in tree[tree.ntips:tree.nnodes - 1]:
#         raise ToytreeError(
#             f"Edge ({edge._idx}, {edge._up._idx if edge._up else None}) "
#             "does not form a quadripartition. Select an internal edge.")

#     # tree will be unrooted, so get edge as collection of leaves, or as
#     # collection of reciprocal set of leaves, if
#     tree = tree.unroot()
#     cache = {}

#     # build a cache of descendants for each Node.
#     for node in tree:

#         # if tip simply store its info
#         if node.is_leaf():
#             cache[node] = {node}
#             continue

#         # get tips from each child, and store union to this node
#         vals = [cache[i] for i in node.children]

#         # cache only the closest Node (exclude_internal_nodes ignored)
#         if contract_partitions:
#             cache[node] = {node}

#         # cache tip Nodes with or without internal Nodes
#         else:
#             # cache only the tip Nodes
#             if not include_internal_nodes:
#                 cache[node] = set.union(*vals)
#             # cache all nodes including internals
#             else:
#                 cache[node] = set.union(*vals) | {node}

#     # get set of all relevant nodes for set diffs below
#     if include_internal_nodes:
#         nodes = set(cache)
#     else:
#         nodes = set(tree.get_nodes()[:tree.ntips])

#     # iterate over internal nodes/edges in the tree
#     for node in tree[tree.ntips: tree.nnodes - 1]:

#         # get >= 2 sets of nodes for children below
#         below = [cache[i] for i in node.children]

#         # get >= 1 sets of nodes to represent >= 1 clades above.
#         above_left = [cache[sis] for sis in node.get_sisters()]

#         # if up is root then above_left contains all remaining clades
#         if node._up.is_root():
#             above = above_left
#             # TODO: need to expand partitions near root?
#         else:
#             # print(node, below, above_left)
#             above_right = nodes - {node, node._up} - set.union(*above_left + below)

#             # represent final clade by all its desc, or by its mrca.
#             if contract_partitions:
#                 above = above_left + [{tree.get_mrca_node(*above_right)}]
#             else:
#                 above = above_left + [above_right]

#         # add others
#         bpair = itertools.combinations(below, 2)
#         apair = itertools.combinations(above, 2)
#         for x1, x2 in itertools.product(bpair, apair):
#             # require all to exclude root in unrooted tree: ()
#             if all([x1[0], x1[1], x2[0], x2[1]]):
#                 if feature is None:
#                     yield ((x1[0], x1[1]), (x2[0], x2[1]))
#                 else:
#                     yield (
#                         tuple({getattr(j, feature) for j in i} for i in x1),
#                         tuple({getattr(j, feature) for j in i} for i in x2),
#                     )


# def _iter_quartet_sets(tree: ToyTree, feature: str = None) -> Iterator[Set]:
#     """Generator to yield all combinations of four tip Nodes.

#     This is simpler than toytree.enumerate.iter_quartets because it
#     does not try to return the resolution of quartets. It is thus
#     also much faster, but of less use.
#     """
#     # select function to format returned quartet set
#     if feature is None:
#         sformat = set
#     else:
#         def sformat(qrt) -> Set[str]:
#             return set(getattr(node, i) for node in qrt)
#     for qrt in itertools.combinations(sorted(tree)[:tree.ntips], 4):
#         yield sformat(qrt)


# def get_bipartitions_table(
#     tree: ToyTree,
#     include_internal_nodes: bool = False,
#     include_singleton_partitions: bool = False,
#     dtype: type = int,
# ) -> pd.DataFrame:
#     """Return a DataFrame with partitions in binary format.

#     Parameters
#     ----------

#     Example
#     -------
#     >>> toytree.rtree.unittree(5).enum.get_bipartitions_table()
#     >>>    r0   r1   r2   r3   r4
#     >>> 0   1    1    0    0    0
#     >>> 1   0    0    0    1    1

#     """
#     bits = list(
#         iter_bipartitions(
#             tree,
#             feature="idx",
#             include_internal_nodes=include_internal_nodes,
#             include_singleton_partitions=include_singleton_partitions,
#         )
#     )
#     cols = tree.nnodes - 1 if include_internal_nodes else tree.ntips
#     arr = np.zeros(shape=(len(bits), cols), dtype=dtype)
#     for idx, bit in enumerate(bits):
#         arr[idx, bit[0]] = 1
#     if include_singleton_partitions:
#         index = range(tree.ntips, tree.ntips + len(bits))
#     else:
#         index = None
#     return pd.DataFrame(arr, columns=tree.get_tip_labels(), index=index)


# @add_subpackage_method(TreeEnumAPI)
# def get_n_quartets(tree: ToyTree) -> int:
#     """Return the number of quartets..."""
#     qiter = itertools.combinations(range(tree.nnodes), 4)
#     return sum(1 for i in qiter)



# if __name__ == "__main__":

#     import toytree

#     BTREE = toytree.rtree.unittree(ntips=5, seed=123)
#     ITREE = toytree.rtree.unittree(ntips=5, seed=321)
#     UTREE = toytree.rtree.unittree(ntips=5, seed=123).unroot()
#     for TREE in (BTREE, ITREE, UTREE):
#         TREE.treenode.draw_ascii()

#         # assign internal node names for viewing results
#         for node in TREE:
#             if not node.is_leaf():
#                 node.name = "-".join(node.get_leaf_names()[::-1])

#         print("\nBIPART SETS-----------")
#         for i in iter_bipartitions(TREE, feature="name", sort=False):
#             print(i)

#         print("\nBIPARTS-----------")
#         for i in iter_bipartitions(TREE, feature='name'):
#             print(i)

#         print("\nBIPARTS W INTERNAL-----------")
#         for i in iter_bipartitions(TREE, feature='name', include_internal_nodes=True):
#             print(i)

#         print("\nBIPARTS w ROOT indicated-----------")
#         for i in iter_bipartitions(TREE, feature='name', indicate_root=True):
#             print(i)

#         print("\nQUARTETS-----------")
#         for i in iter_quartets(TREE, collapse=True):
#             print(i)

#         print("\nQUADRIPART SETS-----------")
#         for i in iter_edge_quadripartition_sets(TREE):
#             print(i)

#         print("\nQUADRIPART SETS CONTRACTED-----------")
#         for i in iter_edge_quadripartition_sets(TREE, contract_partitions=True):
#             print(i)

#         print("\nQUADRIPARTITIONS-----------")
#         for i in iter_quadripartitions(TREE, feature="name", collapse=True):
#             print(i)

#         print("\nBIPART FOR EDGE 1-----------")
#         print(get_bipartition_for_edge(TREE, 1))

#         print("\nBIPART FOR EDGE 5-----------")
#         print(get_bipartition_for_edge(TREE, 5))

#         print("\nBIPART FOR EDGE 5 w INTERNAL-----------")
#         print(get_bipartition_for_edge(TREE, 5, include_internal_nodes=True))

#         print("\nQUARTETS FOR EDGE 1-----------")
#         print(get_quartets_for_edge(TREE, 1))

#         print("\nQUARTETS FOR EDGE 5-----------")
#         print(get_quartets_for_edge(TREE, 5))

#         print("\nQUADPARTS FOR EDGE 5-----------")
#         # print(get_quadripartitions_for_edge(TREE, 5, include_internal_nodes=True))

#         print("===================================================")

#     raise SystemExit()
#     # does every placement of the root yield a different topology?
#     for edge in UTREE:
#         if edge.up:
#             t = UTREE.root(edge)
#             t.treenode.draw_ascii()
#             # print(list(iter_bipartitions(t, 'name', exclude_root=False)))


#     # FIXME: make None work.
#     # for i in iter_bipartitions(TREE, feature=None):
#         # print(i)




#!/usr/bin/env python

"""BaseLayout class from which child classes define other layouts.

This class is used to get the (x,y) coordinates at which to place
nodes on a set of Cartesian coordinate axes when plotting. The layout
types currently supported are 'right', 'left', 'up', 'down', 
'circular' and 'unrooted'. 

Note
----
The circular layout is projected into Cartesian coordinates at the
end, not polar coordinates.
"""

from typing import TypeVar, Sequence, Optional
from abc import ABC
import numpy as np
from loguru import logger
from toytree.style import TreeStyle
from toytree.utils import ToytreeError

ToyTree = TypeVar("ToyTree")
Node = TypeVar("Node")
logger = logger.bind(name="toytree")


class BaseLayout(ABC):
    """Abstract base class for Layout objects.

    Layout class object generates a `.style` object as a TreeStyle
    updated with user kwargs to `.draw()`, and uses relevant args
    in this to build a `.coords` array with Node coordinates, which
    is affected by 'layout', 'use_edge_lengths' and 'tip_labels_align'.
    """

    def __init__(
        self,
        tree: ToyTree,
        style: TreeStyle,
        fixed_order: Optional[Sequence[str]] = None,
        fixed_position: Optional[Sequence[float]] = None,
    ):
        self.tree = tree
        """: ToyTree from which tree shape will be extracted."""
        self.style = style
        """: TreeStyle dict-like from which layout style is extracted."""
        self.fixed_order = fixed_order
        """: List of tip names to override default layout tip order."""
        self.fixed_position = fixed_position
        """: List of tip pos to override default layout tip spacing."""

        # ----------- results to be generated ------------------
        self.coords: np.ndarray = None
        """: Node coordinates in the projected layout."""
        self.tcoords: np.ndarray = None
        """: Tip node coordinates in the projected layout. Note: these
        can be different from those in coords when tip_labels_align=True."""
        self.angles: np.ndarray = None
        """: Tip label angles in the projected layout."""

        # subclasses have a run function that fill .coords and .tcoords
        self.run()

    def run(self):
        """BaseLayout has no run function."""
        raise ToytreeError(f"Use a subclass of {self.__class__}")



#!/usr/bin/env python

"""A simple polygon shape Marker in toyplot style.

SVG
---
<polygon points="100,100 150,25 150,75 200,0" fill="none" stroke="black" />

Python
------
In toytree this code is used only to setup container tree (e.g.,
species trees) drawings. But, in the future this could be ported
to toyplot as a more general polygon shape object for complex
shapes such as maps, etc.

>>> toytree.container(...)
"""

from typing import List, Tuple

import functools
import xml.etree.ElementTree as xml
from multipledispatch import dispatch

import numpy as np
import toyplot.html
from toyplot.mark import Mark
from toytree.color import ToyColor
from toytree.color import concat_style_fix_color


class Polygon(Mark):
    def __init__(self, points: np.ndarray, **kwargs):
        Mark.__init__(self, annotation=kwargs.get("annotation", False))

        self._coordinate_axes: List[str] = ["x", "y"]
        self._table = points
        # self.points: str = self._points_to_str(points)
        self.stroke: str = kwargs.get("stroke", "black")
        self.stroke_width: float = kwargs.get("stroke_width", 2)
        self.stroke_opacity: float = kwargs.get("stroke_opacity", 1)
        self.stroke_opacity: float = kwargs.get("stroke-opacity", self.stroke_opacity)

        self.fill: str = kwargs.get("fill", ToyColor("black"))
        self.fill_opacity: float = kwargs.get("fill_opacity", 0.5)
        self.fill_opacity: float = kwargs.get("fill-opacity", self.fill_opacity)

    def domain(self, axis: str) -> np.ndarray:
        """Return position of Marks on a coordinate axis."""
        index = self._coordinate_axes.index(axis)
        domain = toyplot.data.minimax(self._table[:, index])
        return domain

    def extents(self, axes: List[str]) -> Tuple[np.ndarray, np.ndarray]:
        """Return extent of Marks extending from their coordinates."""
        extents = [np.zeros(self._table.shape[0]) for i in range(4)]
        return tuple(self._table.T), tuple(extents)


# ---------------------------------------------------------------------
# Register multipledispatch to use the toyplot.html namespace
dispatch = functools.partial(dispatch, namespace=toyplot.html._namespace)


# register a _render function for ToyTreeMark objects
@dispatch(toyplot.coordinates.Cartesian, Polygon, toyplot.html.RenderContext)
def _render(axes, mark, context):
    RenderPolygon(axes, mark, context)
# ---------------------------------------------------------------------


class RenderPolygon:
    def __init__(self, axes, mark, context):
        self.axes = axes
        self.mark = mark
        self.context = context

        self.mark_xml: xml.SubElement = xml.SubElement(
            context.parent, "g",
            id=context.get_id(self.mark),
            attrib={"class": "toytree-Polygon"},
        )

        xproj = self.axes.project("x", self.mark._table[:, 0])
        yproj = self.axes.project("y", self.mark._table[:, 1])
        points = " ".join([f"{i[0]},{i[1]}" for i in zip(xproj, yproj)])
        xml.SubElement(
            self.mark_xml, "polygon",
            points=points,
            style=concat_style_fix_color({
                "fill": self.mark.fill,
                "fill-opacity": self.mark.fill_opacity,
                "stroke": self.mark.stroke,
                "stroke-width": self.mark.stroke_width,
                "stroke-opacity": self.mark.stroke_opacity
            }),
        )


if __name__ == "__main__":

    import toytree
    import toyplot

    arr = np.array([
        [0, 0],
        [5, 0],
        [6, 2],
        [3, 3],
        [0, 0],
    ])
    pol = Polygon(points=arr, stroke="blue")
    cvs = toyplot.Canvas(300, 300)
    axe = cvs.cartesian()
    axe.add_mark(pol)

    toytree.utils.show(cvs)



#!/usr/bin/env python

"""

Adds admixture edge to existing ToyTree Object.

Method
------

Creates and draws an admixture edge between two tree tips
>>> annotate.add_admixture_edge()


"""

import numpy as np
import toyplot
from toyplot.coordinates import Cartesian
from loguru import logger
from toytree import ToyTree, Node
from toytree.core.apis import add_subpackage_method, AnnotationAPI
from toytree.annotate.src.checks import get_last_toytree_mark, assert_tree_matches_mark



@add_subpackage_method(AnnotationAPI)
def add_admixture_edge(
    tree: ToyTree,
    axes: Cartesian,
    src: Node = None,
    dest: Node = None
) -> Mark:
    """Returns a Toyplot mark of the admixture edge added.

    Parameters
    ----------

    src: source node at which the admixture edge will start
    dest: destination where the admixture edge will end


    Below is old implementation 
    TODO: add functionality for this argument to add_admixture_edge
    Admixture edges add colored edges to the plot in the 
                style of the 'edge_align_style'. These will be drawn
                from (source, dest, height, width, color). Example:
                `>>> [(4, 3, 50000, 3, 'red')]`.
            fixed_order: Sequence[str]
    """ 


#!/usr/bin/env python

"""This is a fork of ipcoal.msc.get_embedding_table

Example
-------
>>> get_genealogy_embedding_table(sptree, gtree, imap)
"""

from typing import Mapping, Sequence
from loguru import logger
import numpy as np
import pandas as pd
import toytree

logger = logger.bind(name="toytree")
COLUMNS = ['start', 'stop', 'st_node', 'neff', 'nedges', 'dist', 'gidx']


def _get_fast_genealogy_embedding_table(
    species_tree: toytree.ToyTree,
    genealogy: toytree.ToyTree,
    imap: Mapping[str, Sequence[str]],
    gidx: int,
) -> np.ndarray:
    """Internal function to embed a single genealogy.

    This function is used within `get_genealogy_embedding_table()`.
    See the latter function for usage.
    """
    # store temporary results in a dict
    split_data = []
    edge_encode = []

    # dict to update tips to their ancestor if already coalesced.
    name_to_node = {i._name: i for i in genealogy[:genealogy.ntips]}

    # get table of gene tree node heights
    gt_node_heights = np.array([i._height for i in genealogy])

    # iterate over stree from tips to root storing stree node data
    for st_node in species_tree:
        # get all gtree names descended from this sptree interval
        gt_tips = set.union(*(set(imap[i]) for i in st_node.get_leaf_names()))

        # get gtree nodes descended from this interval.
        gt_nodes = set(name_to_node[i] for i in gt_tips)

        # get internal nodes in this TIME interval (coalescences)
        lower = st_node.height + 0.0001  # zero-align bug
        upper = np.inf if st_node.is_root() else st_node._up._height
        nodes_in_time_slice = (gt_node_heights > lower) & (gt_node_heights < upper)

        # filter internal nodes must be ancestors of gt_nodes (in this interval)
        # get internal nodes in this interval by getting all nodes in this
        # time interval and requiring that all their desc are in gt_nodes.
        inodes = []
        for iidx in nodes_in_time_slice.nonzero()[0]:
            gt_node = genealogy[iidx]
            if not (set(gt_node.get_leaf_names()) - gt_tips):
                inodes.append(gt_node)

        # vars to be updated at coal events
        start = st_node._height
        edges = {i._idx for i in gt_nodes}

        # iterate over internal nodes
        for gt_node in sorted(inodes, key=lambda x: x._height):
            # add interval from start to first coal, or coal to next coal
            split_data.append([
                start,
                gt_node._height,
                st_node._idx,
                st_node.Ne,
                len(edges),
                0.,
                gidx,
            ])
            edge_encode.append(sorted(edges))

            # update counters and indexers
            start = gt_node._height
            edges.add(gt_node._idx)
            try:
                for child in gt_node._children:
                    edges.remove(child._idx)
            except KeyError as err:
                raise ValueError("gene tree cannot be embedded in species tree.") from err
            for tip in gt_node.get_leaves():
                name_to_node[tip.name] = gt_node

        # add non-coal interval
        split_data.append([
            start,
            st_node._up._height if st_node._up else np.inf,
            st_node._idx,
            st_node.Ne,
            len(edges),
            0.,
            gidx,
        ])
        edge_encode.append(sorted(edges))

    # to array and calculate dists from start and stop
    arr = np.vstack(split_data)
    arr[:, 5] = arr[:, 1] - arr[:, 0]
    return arr, edge_encode


def get_genealogy_embedding_table(
    species_tree: toytree.ToyTree,
    genealogies: Sequence[toytree.ToyTree],
    imap: Mapping[str, Sequence[str]],
    encode: bool = False,
) -> pd.DataFrame:
    """Return a genealogy embedding table as a human-readable dataframe.

    The embedding table represents intervals in a species tree that
    have piece-wise constant coalescent rates and which are divided
    by coalescent events in the gene tree, or speciation events in
    the species tree.

    Raises ipcoalError if genealogy cannot be embedded in the species
    tree (e.g., coalescence earlier than species divergence allows).

    Parameters
    ----------
    species_tree: ToyTree
        A ToyTree with Ne values mapped to Nodes and edge lengths in
        units of generations.
    genealogies: ToyTree or Sequence[ToyTree]
        One or more ToyTrees representing gene trees embedded in the
        species tree with edge lengths in units of generations.
    imap: Dict[str, List[str]]
        A dict mapping gene tree tip names to species tree tips names.
    encode: bool
        If True genealogy Node IDs are included as one-hot-encoded in
        place of the 'edges' column.

    Examples
    --------
    >>> sptree = toytree.rtree.imbtree(ntips=4, treeheight=1e6)
    >>> model = ipcoal.Model(sptree, Ne=1e5, nsamples=2)
    >>> model.sim_trees(10)
    >>> imap = model.get_imap_dict()
    >>> gtrees = model.df.genealogy
    >>> table = get_genealogy_embedding_table(model.tree, gtrees, imap)
    """
    # get genealogies as Sequence[ToyTree]
    if isinstance(genealogies, (toytree.ToyTree, str)):
        genealogies = [genealogies]
    if isinstance(genealogies[0], str):
        genealogies = toytree.mtree(genealogies).treelist

    # iterate over genealogies
    garrs = []
    earrs = []
    for gidx, gtree in enumerate(genealogies):
        args = (species_tree, gtree, imap, gidx)
        garr, earr = _get_fast_genealogy_embedding_table(*args)
        garrs.append(garr)
        earrs.append(earr)

    # concatenate
    embedding = pd.DataFrame(np.vstack(garrs), columns=COLUMNS)
    embedding[["st_node", "nedges", "gidx"]] = embedding[["st_node", "nedges", "gidx"]].astype(int)

    # return 'edges' column with lists of Node IDs
    if not encode:
        embedding["edges"] = np.array(earrs, dtype=object).flatten()
        return embedding

    # else return one-hot-encoded columns (kinda slow, for didactic)
    nnodes = genealogies[0].nnodes
    encoding = np.zeros((embedding.shape[0], nnodes), dtype=np.uint8)
    i = 0
    for gidx, earr in enumerate(earrs):
        for nodes in earr:
            encoding[i, nodes] = 1
            i += 1
    return pd.concat([embedding, pd.DataFrame(encoding)], axis=1)


if __name__ == "__main__":

    import ipcoal
    SPTREE = toytree.rtree.baltree(2, treeheight=1e6)
    MODEL = ipcoal.Model(SPTREE, Ne=200_000, nsamples=4, seed_trees=123)
    MODEL.sim_trees(2)
    GENEALOGIES = toytree.mtree(MODEL.df.genealogy)
    IMAP = MODEL.get_imap_dict()

    # human readable embedding (2D) mostly for debugging and teaching.
    print(get_genealogy_embedding_table(MODEL.tree, GENEALOGIES, IMAP).head(10))
    print('...')



#!/usr/bin/env python

"""Use toyplot html functions to add markers

"""

from toyplot.html import (_draw_bar, _draw_triangle, _draw_circle, _draw_rect)


def render_marker(marker_xml, marker) -> None:
    """Call the toyplot render function for the selected marker."""
    if marker.shape == "|":
        _draw_bar(marker_xml, marker.size)
    elif marker.shape == "/":
        _draw_bar(marker_xml, marker.size, angle=-45)
    elif marker.shape == "-":
        _draw_bar(marker_xml, marker.size, angle=90)
    elif marker.shape == "\\":
        _draw_bar(marker_xml, marker.size, angle=45)
    elif marker.shape == "+":
        _draw_bar(marker_xml, marker.size)
        _draw_bar(marker_xml, marker.size, angle=90)
    elif marker.shape == "x":
        _draw_bar(marker_xml, marker.size, angle=-45)
        _draw_bar(marker_xml, marker.size, angle=45)
    elif marker.shape == "*":
        _draw_bar(marker_xml, marker.size)
        _draw_bar(marker_xml, marker.size, angle=-60)
        _draw_bar(marker_xml, marker.size, angle=60)
    elif marker.shape == "^":
        _draw_triangle(marker_xml, marker.size)
    elif marker.shape == ">":
        _draw_triangle(marker_xml, marker.size, angle=-90)
    elif marker.shape == "v":
        _draw_triangle(marker_xml, marker.size, angle=180)
    elif marker.shape == "<":
        _draw_triangle(marker_xml, marker.size, angle=90)
    elif marker.shape == "s":
        _draw_rect(marker_xml, marker.size)
    elif marker.shape == "d":
        _draw_rect(marker_xml, marker.size, angle=45)
    elif marker.shape and marker.shape[0] == "r":
        width, height = marker.shape[1:].split("x")
        _draw_rect(
            marker_xml, marker.size,
            width=float(width), height=float(height))
    elif marker.shape == "o":
        _draw_circle(marker_xml, marker.size)
    elif marker.shape == "oo":
        _draw_circle(marker_xml, marker.size)
        _draw_circle(marker_xml, marker.size / 2)
    elif marker.shape == "o|":
        _draw_circle(marker_xml, marker.size)
        _draw_bar(marker_xml, marker.size)
    elif marker.shape == "o/":
        _draw_circle(marker_xml, marker.size)
        _draw_bar(marker_xml, marker.size, -45)
    elif marker.shape == "o-":
        _draw_circle(marker_xml, marker.size)
        _draw_bar(marker_xml, marker.size, 90)
    elif marker.shape == "o\\":
        _draw_circle(marker_xml, marker.size)
        _draw_bar(marker_xml, marker.size, 45)
    elif marker.shape == "o+":
        _draw_circle(marker_xml, marker.size)
        _draw_bar(marker_xml, marker.size)
        _draw_bar(marker_xml, marker.size, 90)
    elif marker.shape == "ox":
        _draw_circle(marker_xml, marker.size)
        _draw_bar(marker_xml, marker.size, -45)
        _draw_bar(marker_xml, marker.size, 45)
    elif marker.shape == "o*":
        _draw_circle(marker_xml, marker.size)
        _draw_bar(marker_xml, marker.size)
        _draw_bar(marker_xml, marker.size, -60)
        _draw_bar(marker_xml, marker.size, 60)



#!/usr/bin/env python

"""Modify ticks and labels on tree time axis into a scale bar.

"""

from typing import Optional
import numpy as np
import toyplot
from toyplot.coordinates import Cartesian
from loguru import logger
from toytree import ToyTree
from toytree.core.apis import add_subpackage_method, AnnotationAPI
from toytree.annotate.src.checks import get_last_toytree_mark, assert_tree_matches_mark


logger = logger.bind(name="toytree")
__all__ = ["add_axes_scale_bar"]


@add_subpackage_method(AnnotationAPI)
def add_axes_scale_bar(
    tree: ToyTree,
    axes: Cartesian,
    ymax: Optional[float] = None,
    only_inside: bool = True,
    # nticks: Optional[int] = None,
    # above/below,
    # padding,margin,near,far...
    # positive/negative time,
) -> Cartesian:
    """Return a toyplot Cartesian object with custom tick marks.

    This gets tick locations first using toyplot.locator.Extended and
    then sets labels on them using toyplot.locator.Explicit, because
    we need time scale bar to be non-negative when axes are rotated
    for trees facing different directions.

    Note
    -----
    This is a convenience function to provide just *one* way to style
    the axes. You can also instead use the `toyplot.locator` module to
    apply many more additional options for styling ticks.
    """
    mark = get_last_toytree_mark(axes)
    assert_tree_matches_mark(tree, mark)

    # the axes is either new or passed as an arg, and the scale_bar
    # arg is True or a (float, int), so we need to style the ticks.
    if mark.layout in ("r", "l"):
        cwidth = axes._xmax_range - axes._xmin_range
        nticks = max((5, np.floor(cwidth / 75).astype(int)))
        axes.show = True
        axes.y.show = False
        axes.x.show = True
        axes.x.ticks.show = True
        left, right = mark.domain('x')
        tree_height = ymax if ymax else right - left

    # ...
    elif mark.layout in ("u", "d"):
        cheight = axes._ymax_range - axes._ymin_range
        nticks = max((5, np.floor(cheight / 75).astype(int)))
        axes.show = True
        axes.x.show = False
        axes.y.show = True
        axes.y.ticks.show = True
        bottom, top = mark.domain('y')
        tree_height = ymax if ymax else top - bottom

    # e.g., unrooted layout with axes shown (e.g., ts='p')
    else:
        # nticks = max((4, np.floor(style.height / 75).astype(int)))
        nticks = 5
        axes.show = True
        axes.x.show = True
        axes.y.show = False

        # TODO: could choose x or y here based on start angle...
        axes.x.ticks.show = True
        left, right = mark.domain('x')
        tree_height = (right - left) / 2

    # get tick locator
    # lct = toyplot.locator.Extended(count=nticks, only_inside=only_inside)
    lct = toyplot.locator.Extended(count=nticks, only_inside=False)

    # get root tree height
    if not tree_height:
        tree_height = 1e-6
    # logger.warning(tree_height)
    if mark.layout in "ru":
        locs = lct.ticks(-tree_height, -0)[0]
        locs = locs[locs >= -tree_height]
    elif mark.layout in "ld":
        locs = lct.ticks(0, tree_height)[0]
        locs = locs[locs <= tree_height]
    else:
        locs = lct.ticks(0, tree_height)[0]
        locs = locs[locs <= tree_height]

    # apply unit scaling
    if mark.scale_bar is False:
        labels = abs(locs.copy())
    elif isinstance(mark.scale_bar, (int, float)):
        labels = abs(locs / mark.scale_bar)
    else:
        labels = abs(locs.copy())
    labels = [np.format_float_positional(i, precision=6, trim="-") for i in labels]

    # set the ticks locator
    if mark.layout in "rl":
        axes.x.ticks.locator = toyplot.locator.Explicit(
            locations=locs + mark.xbaseline,
            labels=labels,
        )
    elif mark.layout in "ud":
        axes.y.ticks.locator = toyplot.locator.Explicit(
            locations=locs + mark.ybaseline,
            labels=labels,
        )
    else:
        axes.x.ticks.locator = toyplot.locator.Explicit(
            locations=locs + mark.xbaseline,
            labels=labels,
        )
        axes.x.domain_min = 0
    # print(locs, labels)
    return axes


if __name__ == "__main__":

    import toytree
    orig = toytree.rtree.rtree(5, seed=123)
    orig.set_node_data("orig", {i: i.idx for i in orig}, inplace=True)
    # orig.draw('p', node_labels="orig");
    a, b = orig.mod.bisect(1)
    c, ax, m = a.draw()
    add_axes_scale_bar(a, ax, )#scale_bar=True);



#!/usr/bin/env python

"""In development...

A CloudTreeMark is a Mark containing many overlapping trees. To reduce
the html+css size we use a custom render function?
"""

from typing import Sequence, TypeVar
from loguru import logger
# from toytree import MultiTree
from toytree.drawing import ToyTreeMark
# from toytree.core import Canvas, Cartesian
from toytree.style import tree_style_to_css_dict
from toytree.drawing.src.draw_toytree import (
    get_layout,
    get_tree_style_updated_by_draw_args,
)
# from toytree.drawing.src.mark_cloudtree import CloudTreeMark

Mark = TypeVar("Mark")
MultiTree = TypeVar("MultiTree")
logger = logger.bind(name="toytree")


def draw_cloudtree(mtree: MultiTree, **kwargs) -> Sequence[Mark]:
    """Parse arguments to draw_cloudtree and return drawing objects.

    CloudTree is a Mark similar to a ToyTree but with many overlapping
    sets of edges, where each set can be individually styled. Only one
    set of tip labels is plotted.
    """
    # which trees to plot
    idxs = kwargs.get("idxs")
    if idxs:
        mtree = mtree.copy()
        mtree.treelist = mtree.treelist[idxs]

    # TODO: allow jitter options along tip spread axis
    jitter = kwargs.get("jitter", 0.)

    # get fixed order of tips from consensus tree if not provided.
    fixed_order = kwargs.get("fixed_order", None)
    if fixed_order in [True, False, None]:
        fixed_order = (
            mtree
            .get_consensus_tree()
            .get_tip_labels()
        )

    # TODO
    else:
        # require that the user fixed_order argument includes all of
        # the tips present across all the trees.
        # assert len(fixed_order) == max_num_tips, "..."
        pass

    # get a dict mapping tips to indices for the full set of tips
    # tip_pos = dict(zip(fixed_order, range(len(fixed_order))))

    # iterate over trees and for each, allow the trees individual style
    # to override any styles not fixed or from kwargs.
    marks = []
    for tidx, tree in enumerate(mtree):

        # only show tips for the first tree
        if tidx:
            kwargs["tip_labels"] = False

        # hard-coded to disallow some styles
        kwargs["scale_bar"] = False

        # get the tree's style
        style = get_tree_style_updated_by_draw_args(tree, **kwargs)

        # using fixed order while allowing diff numbers of tips
        # style.fixed_order = [tip_pos[i] for i in tree.get_tip_labels()]

        # override styles not set by user
        style.edge_type = kwargs.get("edge_type", 'c')
        if style.edge_style.stroke_opacity is None:
            style.edge_style.stroke_opacity = 1 / len(mtree) * 3

        # get coordinate layout of this tree
        layout = get_layout(
            tree=tree,
            style=style,
            fixed_order=fixed_order,
            fixed_position=kwargs.get("fixed_position", None),
            interior_algorithm=kwargs.get("interior_algorithm", 1),
        )

        if style.tip_labels_angles is None:
            style.tip_labels_angles = layout.angles

        mark = ToyTreeMark(
            ntable=layout.coords,
            ttable=layout.tcoords,
            etable=tree.get_edges('idx'),
            **tree_style_to_css_dict(style),
        )
        marks.append(mark)

    # get a Layout with coordinates projected based on the first tree
    return marks


if __name__ == "__main__":

    import toytree
    import toyplot
    trees = [toytree.rtree.coaltree(k=6, seed=i) for i in range(100)]
    mtree = toytree.mtree(trees)
    c, a, m = mtree.draw_cloud_tree()
    toytree.utils.show([c], tmpdir="~")
    # canvas = toyplot.Canvas()
    # axes = canvas.cartesian()
    # kwargs = dict(
    #     tree_style=None, axes=axes, kwargs={}, edge_widths=3,
    #     fixed_position=None,
    #     # fixed_order=["r1", "r0", "r3", "r4", "r5", "r2"],
    # )

    # mtree[10].style.edge_style.stroke = "red"
    # mtree[10].style.edge_style.stroke_opacity = 1
    # marks = draw_cloudtree(mtree, **kwargs)
    # if marks:
    #     toytree.utils.show([canvas], tmpdir="~")



#!/usr/bin/env python

"""Custom toyplot.Mark for displaying pie charts.
"""

import numpy as np
import toyplot
from toyplot.mark import Mark


class PieChartMark(Mark):
    """Return a custom Mark as a scatterplot-like with Pie markers.

    To create PieChart instances a user should use the standard
    :meth:`draw` function of ToyTree instances entering an ndarray
    with (ncategories, nnodes) dimension; or, enter the same data
    as an argument to the :meth:tree.annotate.pie_charts() function.

    Parameters
    ----------
    ntable: numpy.ndarray
        Coordinates of markers
    data: numpy.ndarray
        Percentages.
    annotation: bool
        Mark is an 'annotation' not affecting data domain.

    Examples
    --------
    >>> tree = toytree.rtree.unittree(ntips=10, seed=123)
    >>> pie_data = np.repeat([1/3, 1/3, 1/3], tree.nnodes)
    >>> c, a, m = tree.draw(node_colors=pie_data)

    >>> tree = toytree.rtree.unittree(ntips=10, seed=123)
    >>> c, a, m = tree.draw(node_colors=pie_data)
    >>> tree.annotate.pie_charts(axes, mark, ...)
    """
    def __init__(
        self,
        coordinates,
        data,
        sizes,
        colors,
        istroke,
        istroke_width,
        ostroke,
        ostroke_width,
        rotate,
        xshift,
        yshift,
    ):
        Mark.__init__(self, annotation=True)
        self._coordinate_axes = ['x', 'y']
        self.coordinates = np.array(coordinates)
        self.data = data
        self.sizes = toyplot.broadcast.scalar(sizes, self.coordinates.shape[0])
        self.ostroke = ostroke
        self.ostroke_width = ostroke_width
        self.istroke = istroke
        self.istroke_width = istroke_width
        self.rotate = rotate
        self.colors = colors
        self.xshift = xshift
        self.yshift = yshift

        # radius is half the size
        self.sizes /= 2.

    def domain(self, axis):
        """The domain of data to fit on Canvas."""
        index = self._coordinate_axes.index(axis)
        domain = toyplot.data.minimax(self.coordinates[:, index])
        return domain

    def extents(self, axes):
        """..."""
        node_extents = [self.sizes / 2. for i in range(4)]
        node_extents[0] *= -1
        node_extents[2] *= -1
        node_extents[0] += self.xshift
        node_extents[1] += self.xshift
        node_extents[2] += self.yshift
        node_extents[3] += self.yshift
        coords = (
            self.coordinates[:, 0].copy(),
            self.coordinates[:, 1],
        )
        return coords, node_extents


if __name__ == "__main__":

    pass
    # import toyplot
    # c = toyplot.Canvas()
    # a = c.cartesian()
    # mark = PieChartMark()






"""TODO.

"""



# validate_edge_widths(tree, 0)
# validate_edge_widths(tree, 1)
# validate_edge_widths(tree, 1.555)
# validate_edge_widths(tree, np.arange(tree.nnodes))
# validate_edge_widths(tree, ("dist", ))
# validate_edge_widths(tree, ("dist", 2))
# validate_edge_widths(tree, ("dist", 2, 4))
# validate_edge_widths(tree, ("dist", 2, 4, 0))


#!/usr/bin/env python

"""...

"""

import functools
import xml.etree.ElementTree as xml

from multipledispatch import dispatch
import toyplot.html
from toyplot.coordinates import Cartesian
from loguru import logger

from toytree.drawing.src.render_marker import render_marker
from toytree.color.src.concat import concat_style_fix_color
from toytree.drawing.src.mark_annotation import (
    AnnotationMarker, AnnotationRect)

logger = logger.bind(name="toytree")

# ---------------------------------------------------------------------
# Register multipledispatch to use the toyplot.html namespace
dispatch = functools.partial(dispatch, namespace=toyplot.html._namespace)


# noqa: F811: multidispatch allows reuse the _render() function name
@dispatch(toyplot.coordinates.Cartesian, AnnotationMarker, toyplot.html.RenderContext)
def _render(axes, mark, context):
    render_markers(axes, mark, context)


@dispatch(toyplot.coordinates.Cartesian, AnnotationRect, toyplot.html.RenderContext)
def _render(axes, mark, context):
    render_rect(axes, mark, context)
# ---------------------------------------------------------------------


def render_markers(
    axes: Cartesian,
    mark: AnnotationMarker,
    context: toyplot.html.RenderContext,
) -> None:
    """...

    """
    # create a <g ..> to group markers into.
    axml = xml.SubElement(
        context.parent, "g",
        id=context.get_id(mark),
        attrib={"class": "toytree-Annotation-Markers"},
        style=concat_style_fix_color(mark.style),
    )

    # project coordinates from data units to px units
    nodes_x = axes.project('x', mark.ntable[:, 0])
    nodes_y = axes.project('y', mark.ntable[:, 1])

    # create a marker for each point and add to axml group
    for nidx in range(mark.ntable.shape[0]):

        # create the <marker ...>
        marker = toyplot.marker.create(
            shape=mark.shapes[nidx],
            size=mark.sizes[nidx],
        )

        # create a <g ..> element for marker w/ or w/o unique style
        attrib = {"id": f"Mark-{nidx}"}
        if (mark.colors is None) and (mark.opacity is None):
            marker_xml = xml.SubElement(axml, "g", attrib=attrib)
        elif mark.colors is None:
            style = f"fill-opacity: {mark.opacity[nidx]:.3f}"
            marker_xml = xml.SubElement(axml, "g", attrib=attrib, style=style)
        else:
            style = concat_style_fix_color({
                "fill": None if mark.colors is None else mark.colors[nidx],
                "fill-opacity": None if mark.opacity is None else mark.opacity[nidx],
            })
            marker_xml = xml.SubElement(axml, "g", attrib=attrib, style=style)

        # project marker in coordinate space
        transform = "translate({:.6g},{:.6g})".format(
            nodes_x[nidx] + mark.xshift,
            nodes_y[nidx] + mark.yshift,
        )
        marker_xml.set("transform", transform)

        # get shape from toyplot marker library
        render_marker(marker_xml, marker)


def render_rect(
    axes: Cartesian,
    mark: AnnotationRect,
    context: toyplot.html.RenderContext,
) -> None:
    """Render rectangle Mark and append to HTML.

    custom rect render function to allow transform in px units and
    curved edge styling.
    """
    # create a <g ..> to group markers into.
    axml = xml.SubElement(
        context.parent, "g",
        id=context.get_id(mark),
        attrib={"class": "toytree-Annotation-Rect"},
        style=concat_style_fix_color(mark.style),
    )

    # project coordinates from data units to px units
    left_x = axes.project('x', mark.xtable[:, 0])
    right_x = axes.project('x', mark.xtable[:, 1])
    top_y = axes.project('y', mark.ytable[:, 0])
    bot_y = axes.project('y', mark.ytable[:, 1])

    widths = right_x - left_x
    heights = abs(top_y - bot_y)

    # create a marker for each point and add to axml group
    for nidx in range(mark.ntable.shape[0]):

        # create a <g ..> element for marker w/ or w/o unique style
        attrib = {"id": f"Rect-{nidx}"}
        if (mark.colors is None) and (mark.opacity is None):
            marker_xml = xml.SubElement(axml, "g", attrib=attrib)
        elif mark.colors is None:
            style = f"fill-opacity: {mark.opacity[nidx]:.3f}"
            marker_xml = xml.SubElement(axml, "g", attrib=attrib, style=style)
        else:
            style = concat_style_fix_color({
                "fill": None if mark.colors is None else mark.colors[nidx],
                "fill-opacity": None if mark.opacity is None else mark.opacity[nidx],
            })
            marker_xml = xml.SubElement(axml, "g", attrib=attrib, style=style)

        # position marker in coordinate space
        xpos = left_x[nidx] + mark.xshift
        ypos = bot_y[nidx] - mark.yshift
        transform = f"translate({xpos:.8g},{ypos:.8g})"
        marker_xml.set("transform", transform)

        # add rect to marker: <rect width=x height=y rx=[0-50] ry=[0-50] />
        kwargs = dict(
            width=f"{widths[nidx]:.8g}",
            height=f"{heights[nidx]:.8g}",
            rx="1",
            ry="1",
        )
        _ = xml.SubElement(marker_xml, "rect", **kwargs)


if __name__ == "__main__":

    import toytree

    # base tree drawing
    tree = toytree.rtree.unittree(15)
    c, a, m = tree.draw(layout='d', scale_bar=True, node_sizes=5, node_colors='black')

    m1 = tree.annotate.add_node_bars(
        a,
        bar_min=tree.get_node_data("height").values * 0.8,
        bar_max=tree.get_node_data("height").values * 1.25,
        size=0.3,
        style={"fill-opacity": 0.4, "stroke": None},
        z_index=-1,
        color='red',
    )
    toytree.utils.show(c)



#!/usr/bin/env python

"""Tree enumeration subpackage.


"""

from toytree.enum.src.edges import *
from toytree.enum.src.bipartitions import *
from toytree.enum.src.quadripartitions import *
from toytree.enum.src.quartets import *
from toytree.enum.src.counting import *



#!/usr/bin/env python

"""...

"""

from typing import Tuple
import numpy as np
from loguru import logger
from toytree.layout.src.layout_base import BaseLayout


logger = logger.bind(name="toytree")


class CircularLayout(BaseLayout):
    """Layout for circular tree projection."""

    def run(self):
        """Fills .coords array with x, y values."""
        self.coords = np.zeros(shape=(self.tree.nnodes, 2))
        self.tcoords = self.coords[:self.tree.ntips, :].copy()
        self.set_fan_coords()

    def set_fan_coords(self):
        """Return array with x, y Node coordinates."""

        # position of the *aligned* tips on the fan circumference, the
        # first tips will be at 'start' (e.g., 0) and the final will be
        # at 'end' (e.g., 360 - unit) where unit is space between tips.
        start, end = self._get_start_and_end_angles()
        endpoint = end - start != 360

        # get angles equally spaced from start to end
        angles = np.linspace(start, end, self.tree.ntips, endpoint=endpoint)
        self.angles = angles
        radians = dict(zip(range(self.tree.ntips), np.deg2rad(angles)))

        # the root Node is at position (0, 0), plus any offset style
        hub = self.style.xbaseline, self.style.ybaseline
        self.coords[self.tree.treenode.idx, :] = hub

        # tip node positions will be on a straight lines (spokes) from
        # the root hub to the circum_points.
        for idx in range(self.tree.ntips):
            node = self.tree[idx]
            theta = radians[idx]
            hypo = self.tree.treenode.height - node.height
            delta_y = hypo * np.sin(theta)
            delta_x = hypo * np.cos(theta)
            self.coords[idx, :] = (hub[0] + delta_x, hub[1] + delta_y)
            # set ttable tip heights
            hypo = self.tree.treenode.height
            delta_y = hypo * np.sin(theta)
            delta_x = hypo * np.cos(theta)
            self.tcoords[idx, :] = (hub[0] + delta_x, hub[1] + delta_y)

        # internal node positions are on a spoke pointing towards a
        # circumferal position intermediate between its children's.
        for idx in range(self.tree.ntips, self.tree.nnodes - 1):
            node = self.tree[idx]
            hypo = self.tree.treenode.height - node.height
            theta = np.mean([radians[i.idx] for i in node.children])
            radians[idx] = theta
            delta_y = hypo * np.sin(theta)
            delta_x = hypo * np.cos(theta)
            self.coords[idx, :] = (hub[0] + delta_x, hub[1] + delta_y)
            # node.theta = theta

        # coords of tips around a circumference
        # elif self.mark.layout[0] == 'c':
        #     self.tips_x = np.zeros(ntips)
        #     self.tips_y = np.zeros(ntips)
        #     for idx, angle in enumerate(self.mark.tip_labels_angles):
        #         radian = np.deg2rad(angle)
        #         cordx = 0 + max(self.mark.radii) * np.cos(radian)
        #         cordy = 0 + max(self.mark.radii) * np.sin(radian)
        #         self.tips_x[idx] = self.axes.project('x', cordx)
        #         self.tips_y[idx] = self.axes.project('y', cordy)
        # self.ttable = ....

    def _get_start_and_end_angles(self) -> Tuple[int, int]:
        """Return a tuple with start and end angles as ints.

        Users can enter style as 'c', 'c90', or 'c0-90'. All values
        will be converted to range such that the first is in 0-359,
        and the second is in 1-719.
        - 90,180  -> 90,180
        - 300,100 -> 300,460
        - 359,340 -> 359,609
        - 359,359 -> 359,719  # equivalent to 0-360 starting at -1.
        """
        angles = str(self.style.layout[1:]).strip()

        # if None then use 0-360, if 1 then 0-int, else int-int.
        if not angles:
            start, end = 0, 360
        elif "-" not in angles:
            start, end = 0, int(angles)
        else:
            start, end = (int(i) for i in angles.split("-"))

        # ...
        msg = "circular style malformed. Should be, e.g., 'c', 'c90', 'c0-180'"
        while start < 0:
            start += 360
        while end < start:
            end += 360
        assert end > start, msg
        if end - start > 360:
            end = start + 359
        logger.debug(f"{start}-{end}")
        return start, end

    # def get_radial_coords(self, use_edge_lengths=True):
    #     """
    #     Assign .edges and .verts for node positions in a fan tree.
    #     The farthest tip aligns at the circumference.
    #     """
    #     circ = Circle(self.ttree)
    #     verts = np.zeros((self.ttree.nnodes, 2), dtype=float)

    #     # shortname
    #     if not use_edge_lengths:
    #         nbits = self.ttree.treenode.get_farthest_leaf(True)[1]

    #     # use cache to fill edges array
    #     for idx in range(self.ttree.nnodes):
    #         node = self.ttree[idx]

    #         # leaves: x positions are evenly spaced around circumference
    #         if node.is_leaf() and (not node.is_root()):

    #             # store radians (how far around from zero to 2pi)
    #             node.radians = circ.tip_radians[idx]

    #             # get positions of tips using radians and radius
    #             if use_edge_lengths:
    #                 node.radius = circ.radius - node.height
    #                 node.x, node.y = circ.get_node_coords(node)
    #             else:
    #                 node.radius = nbits
    #                 node.x, node.y = circ.get_node_coords(node)

    #         # internal nodes comes after tips. Inherit position from children.
    #         else:

    #             # height is either distance or nodes from root
    #             if use_edge_lengths:
    #                 node.radius = circ.radius - node.height
    #             else:
    #                 node.radius = sum(1 for i in node.iter_ancestors())
    #                 # max([i.radius for i in node.children]) - 1

    #             # x position is halfway between children x-pos
    #             node.radians = np.mean([i.radians for i in node.children])
    #             node.x, node.y = circ.get_node_coords(node)

    #         # store the x,y vertex positions
    #         verts[node.idx] = [node.x, node.y]
    #     return verts


if __name__ == "__main__":

    import toyplot
    import toytree
    from toytree.drawing.src.draw_toytree import get_tree_style_base, get_layout
    toytree.set_log_level("DEBUG")

    tre = toytree.rtree.unittree(8)
    sty = tre.style
    sty.layout = 'c'
    lay = CircularLayout(tre, sty)
    print(lay.coords)

    canvases = []
    for lay in ["c0-360", "c90-180", "c0-90", "c0-180", "c180-0"]:
        sty = get_tree_style_base(tre, tree_style='r')
        sty.layout = lay
        sty.tip_labels = False
        LAY = get_layout(tre, sty)

        canvas = toyplot.Canvas(400, 300)
        axes = canvas.cartesian()
        axes.graph(
            tre.get_edges(),
            vcoordinates=LAY.coords,
            # vsize=16,
            estyle={"stroke-width": 2},
        )
        canvases.append(canvas)

    toytree.utils.show(canvases)



#!/usr/bin/env python

"""...

"""

from typing import TypeVar
import numpy as np
import toyplot
# from toytree.utils import ToytreeError
from toytree.color import ToyColor
from toytree.style.src.style_base import NodeLabelStyle, TreeStyle


ToyTree = TypeVar("ToyTree")
Color = TypeVar("Color", str, tuple, np.ndarray)

__all__ = [
    "validate_node_labels_style",
]


def validate_node_labels_style(
    tree: ToyTree,
    style: NodeLabelStyle = None,
    **kwargs,
) -> TreeStyle:
    """Ensure tip labels are in px sizes and fill color as ToyColor.

    Parameters
    ----------
    base: TreeStyle or None
        A base tree style to modify. If None then new NodeLabelStyle is used.
    style: Mapping
        A dict of style changes.
    copy: bool
        If False the base is copied and modified, else modified directly.
    """
    if not isinstance(style, NodeLabelStyle):
        style = NodeLabelStyle()

    # update default style with user style args
    for key, val in kwargs.items():
        style[key] = val

    # ensure font-size is in px units
    size = toyplot.units.convert(value=style.font_size, target="px", default="px")
    style.font_size = f"{size:.12g}px"

    # convert fill to a ToyColor
    style.fill = ToyColor(style.fill)
    if style.fill_opacity is None:
        style.fill_opacity = 1.0
    return style


if __name__ == "__main__":

    import toytree

    tree = toytree.rtree.unittree(5)
    style = tree.style.copy()

    print(style.node_labels_style)
    s = validate_node_labels_style(tree, style.node_labels_style, {"-toyplot-anchor-shift": 10}, copy=True)

    print(s)



#!/usr/bin/env python

"""Maxcut algorithm.

TODO
----
- see MQLib for new heuristics to solve max-cut given the graph,
which we construct here.

This is an implementation of the QMC (quartet maxcut) algorithm of
Snir and Rao for fast inference of a supertree from weighted quartets.

The smallest informative piece of phylogenetic information is a quartet
tree (quartet), which is a tree with four tips separated by a single
internal edge. A quartet can be written in newick as "((a,b),(c,d));"
or abbreviated as "ab|cd". Note that this does not mean that 'a' is
more closely related to 'b' than to 'c', only that there is an edge
separating them.

The goal of a supertree construction from quartet trees, or quartet
amalgamation, is to find the tree that is concordant with the greatest
number of input quartet trees. Finding this tree is a NP-hard problem
(Steel 1992).

Development notes
-----------------
- See docs/consensus2.ipynb notebook for current embedding implementation
- For small ntaxa the non-embedding maxcut search is working.
- Have not yet implemented building the tree from the finished set of cuts.

References
----------
The first quartet inference paper
- Snir & Rao 2012 "Quartet MaxCut: A fast algorithm for amalgamating quartet trees".
  https://doi.org/10.1016/j.ympev.2011.06.021

A paper that describes the embedding approach in more detail.
- Snir & Rao 2006 "Using Max Cut to Enhance Rooted Trees Consistency"
  IEEE/ACM Transactions on Computational Biology and Bioinformatics,
  doi: 10.1109/TCBB.2006.58.

A paper that describes the maxcut algorithm in more detail.
- Snir & Rao 2010 "Quartets MaxCut: A Divide and Conquer Quartets Algorithm".
  IEEE/ACM Transactions on Computational Biology and Bioinformatics,
"""

from typing import Union
import itertools
from collections import defaultdict
import numpy as np
import toyplot
from toytree.core import ToyTree, Node
from toytree.core.multitree import MultiTree
from loguru import logger
logger = logger.bind(name="toytree")

# __all__ = [
#     # "get_supertree_from_quartets_by_qmc"
#     # "get_supertree_from_quartets_by_qmc"
#     # "get_quartet_graph",
#     # "get_maxcut_tree_from_quartet_sets",
#     # "get_maxcut_tree_from_multitree",
# ]


##################################################################

def get_weighted_quartets_from_trees(
    trees: Union[MultiTree, list[ToyTree]],
    normalize: bool = False,
) -> dict[tuple[frozenset,frozenset],float]:
    """Return {quartet: weight, ...} from a sequence of trees.

    The weight of each quartet is calculated as the number or
    proportion of trees that contain that quartet.

    Parameters
    ----------
    trees: MultiTree | Sequence[ToyTree]
        An input set of one or more ToyTree objects.
    normalize_by_ntrees: bool
        If True weights are returned as a frequency instead of a count.
    """
    # build weighted graph from quartets
    qrt_counts = {}
    for tree in trees:
        for qrt in tree.enum.iter_quartets(type=frozenset, sort=True):
            if qrt not in qrt_counts:
                qrt_counts[qrt] = 1
            else:
                qrt_counts[qrt] += 1

    # get weights as a proportion.
    if normalize:
        for qrt in qrt_counts:
            qrt_counts[qrt] = qrt_counts[qrt] / len(trees)
    return qrt_counts


##################################################################

def get_graph_from_quartets_qmc(
    quartets: dict[tuple[frozenset,frozenset],float],
    use_weights: bool = False,
) -> dict[frozenset, list[int, int]]:
    """Return {frozenset: weight} for each edge in a graph built from
    quartets where weight is the mean of the quartets that included
    each edge as a partition.

    - good edges are assigned a uniform positive weight (1)
    - bad edges are assigned a -alpha parameter weight.

    Parameters
    ----------
    use_weights: ...
    """
    if use_weights:
        raise NotImplementedError("TODO")
    graph = {}
    # iterate over quartets
    for qrt in quartets:
        # set edge weights between tips on different sides of the quartet (good edges)
        for p1, p2 in itertools.product(*qrt):
            edge = frozenset((p1, p2))
            if edge not in graph:
                graph[edge] = [1, 0]
            else:
                graph[edge][0] += 1

    # set edge weights between tips on same side of the quartet (bad edges)
    for qrt in quartets:
        for part in qrt:
            if part not in graph:
                graph[part] = [0, 1]
            else:
                graph[part][1] += 1
    return graph


##################################################################
# Finding the best cut (partition) for a set of nodes
#
# 1. Compute (plus, minus) score for each edge from quartet sets
# 2a. For small ntaxa test all possible bipartitions.
# 2b. For large ntaxa perform heuristic search to get initial cut (below)
# 3. Distribute nodes randomly on a sphere.
# 4. Move nodes towards center of mass by score/dist weights, iteratively
# 5. Partition on random hyperplane through origin
# 6. Perform greedy search to propose +/- node swap from current
##################################################################

def assign_random_points_on_sphere(ntips: int) -> np.ndarray:
    """Return an array of ntips randomly assigned on a sphere.
    """
    points = []
    for _ in range(ntips):
        # generate random angles
        theta = np.random.uniform(0, 2 * np.pi)  # azimuthal angle
        phi = np.random.uniform(0, np.pi)        # polar angle

        # convert spherical coordinates to Cartesian coordinates
        x = np.sin(phi) * np.cos(theta)
        y = np.sin(phi) * np.sin(theta)
        z = np.cos(phi)
        points.append((x, y, z))
    return np.array(points)


def get_euclidean_dist_arr(points: np.ndarray) -> np.ndarray:
    """Return array of pairwise Euclidean distances between points in 3D space.

    Parameters
    ----------
    points: numpy.ndarray
        A 2D array where each row represents a point (x, y, z).
    """
    # Calculate differences in each dimension
    diffs = points[:, np.newaxis, :] - points[np.newaxis, :, :]

    # Calculate the squared distances
    squared_diffs = diffs ** 2

    # Sum squared differences along the last axis and take the square root
    distances = np.sqrt(np.sum(squared_diffs, axis=-1))
    return distances


def compute_center_of_mass(focal_point, other_points, weights):
    """Return center of mass for each point.

    The center of mass (COM) point of a vertex is the closest point to
    all its neighbors, proportional to the edge weight to each neighbor.
    For example,if a vertex has two neighbors with equal edge weights, 
    its COM is in the middle of the line connecting these two neighbors,
    and if one edge weight is twice the weight of the other, the COM is 
    on the (1/3, 2/3) point on that line.

    Here, a node should move towards those with negative distances
    (bad edge * alpha) and away from those with positive distances
    (good edges). Our goal is to maximize the sum distance score.
    """
    # Calculate the weighted sum of the neighboring points
    weighted_sum = np.sum(weights[:, np.newaxis] * other_points, axis=0)
    
    # Normalize the weighted sum by the total weight (absolute sum of weights to handle negative weights)
    total_weight = np.sum(np.abs(weights))    
    if total_weight == 0:
        return focal_point  # If total weight is zero, return the focal point as the COM
    com_vector = weighted_sum / total_weight
    
    # The center of mass may not be on the unit sphere, so normalize it
    # (Note: I'm not positive this is required)
    # return com_vector
    com_unit_vector = com_vector / np.linalg.norm(com_vector)
    return com_unit_vector


def get_centers_of_mass(points, edges, adists) -> tuple[np.ndarray, np.ndarray]:
    """...

    """
    # iterate over points
    coms = []
    comdists = []
    for idx in range(points.shape[0]):
        # get which neighbors are connected to this one, e.g., [0, 1, 2, 5]
        point = points[idx]
        idxs = np.nonzero((edges[:, 0] == idx) | (edges[:, 1] == idx))[0]
        neis = sorted(set(edges[idxs].flatten()) - {idx})

        # get their coordinates
        coords = points[neis]

        # get their weights
        weights = adists[idxs].copy()

        # how to handle negative weights? Set negatives to zero.
        #weights[weights < 0] = 0.
        if not weights.sum():
            weights[:] = 1.
        else:
            weights = weights / weights.max()
            weights = weights * -1

        # get center of mass
        com = compute_center_of_mass(point, coords, weights)

        # get distance of point from its center of mass
        comdist = np.linalg.norm(com - point)
        coms.append(com)
        comdists.append(comdist)

    # Sample index weighted by its relative comdist
    coms = np.array(coms)
    comdists = np.array(comdists)
    return coms, comdists


def sample_point_far_from_com(points, coms, comdists) -> tuple[int, tuple[float,float,float]]:
    """Return the index of the sampled point far from the COM.
    # Samples probabilitistically weighted by relative distance to COM.

    Move a point towards its center of mass only if it increases the
    sum dist score (greedy algorithm).

    Parameters
    -----------
    ...
    """
    idx = np.random.choice(range(points.shape[0]), p=comdists / sum(comdists))
    return idx, coms[idx]


def move_point_by_distance_on_sphere(point, distance):
    """Move a point by a given distance in a random direction a unit sphere.
    """
    # Generate a random direction on the sphere by creating a random unit vector
    random_direction = np.random.randn(3)
    random_direction /= np.linalg.norm(random_direction)
    
    # Use spherical linear interpolation (slerp) to move the point
    new_point = np.cos(distance) * point + np.sin(distance) * random_direction
    
    # Ensure the result is still on the unit sphere by normalizing it
    new_point /= np.linalg.norm(new_point)
    return new_point


def get_coordinate_shifted_towards_com(
    point: tuple[float,float,float],
    com: tuple[float,float,float],
    step: float,
) -> tuple[float, float, float]:
    """Return coordinate of point2 shifted closer to point1 on the unit sphere by a specified step.

    Parameters
    ----------
    point1: ndarray
        A 3D coordinate on the unit sphere.
    point2: ndarray
        A 3D coordinate on the unit sphere to be shifted.
    step: float
        A value between 0 and 1 indicating how far to shift point2 towards point1.
    """
    # Ensure the points are numpy arrays
    point1 = np.array(com)
    point2 = np.array(point)

    # Calculate the vector from point2 to point1
    direction = point1 - point2
    
    # Calculate the shift amount
    shift_amount = step * np.linalg.norm(direction)

    # Move point2 towards point1
    new_position = point2 + direction / np.linalg.norm(direction) * shift_amount

    # Normalize the new position to ensure it's on the unit sphere
    new_position = new_position / np.linalg.norm(new_position)
    return new_position


def get_cut_from_heuristic_sdp_embedding(graph, alpha, seed, iters):
    """...

    The goal is to embed points on the unit sphere to maximize the
    function (sumdists(good_edges) - (alpha * sumdists(bad_edges))),
    using Euclidean distances between nodes on the sphere. This is
    similar to an SDP algorithm and is referred to as a heuristic
    SDP-like approach by Snir and Rao.

    Given their initial random positions and weights (default 1) move
    nodes iteratively towards their center of mass until a convergence
    is observed.
    """
    # get alphanumeric sorted nodes
    nodes = sorted(frozenset.union(*graph))
    # assign nodes to random points on a 3D sphere to start
    points = assign_random_points_on_sphere(len(nodes))

    # get sumdist of alpha weighted edges
    proposed = points
    edges, alpha_dists = get_edges_and_alpha_distances(nodes, graph, points, alpha)    
    sumdist = last_sum = sum(alpha_dists)

    # optimize: move points locally to optimize sumdist for current
    # alpha until there is no more improvement.
    # for i in range(iters):
    while 1:

        # sample a proposed shift
        com = get_center_of_mass(points, edges, alpha_dists)
        idx = sample_point_far_from_com(com, points)
        pos = get_coordinate_shifted_towards_com(com, points[idx], 0.05)

        # make an array copy with proposed shift points
        proposed = points.copy()
        proposed[idx] = pos

        # calculate score of proposal
        edges, alpha_dists = get_edges_and_alpha_distances(nodes, graph, proposed, alpha)
        sumdist = sum(alpha_dists)

        # accept or reject proposal based on its score
        if sumdist > last_sum:
            points = proposed
            logger.info(f"accepted shift, score = {sumdist:.5f}")
            last_sum = sumdist
    return points


def get_edges_and_alpha_distances(nodes, graph, points, alpha):
    """..."""
    # calculate euclidean distances    
    dists = get_euclidean_dist_arr(points)
    # get edges and weights as arrays using current alpha and Euclidean
    # dists to sum edges as (good_dists - bad_dists * alpha)
    dgraph = graph.copy()
    for edge, (good, bad) in graph.items():
        e = tuple(edge)
        idxs = nodes.index(e[0]), nodes.index(e[1])
        edist = dists[*idxs]
        dgraph[edge] = (edist * good) - (alpha * (edist * bad))
    edges = np.array([tuple(nodes.index(i) for i in j) for j in graph])
    weights = np.array([dgraph[frozenset(i)] for i in dgraph])
    return edges, weights    


def optimize_sphere_embedding(points, edges, weights): # nodes, graph, points, dists, alpha):
    """...

    """
    # calculate center of mass (COM)
    com = get_center_of_mass(points, edges, weights)

    # find who is farthest from COM
    idx = get_index_of_furthest_point_from_com(com, points)

    # get position 50% between that point and COM, but on the unit sphere.
    new_coordinates_on_sphere = get_coordinate_shifted_towards_com(com, points[idx], 0.5)
    points[idx] = new_coordinates_on_sphere
    return points
    # repeat...




def temp():
    # create graph copy w/ distances {{'a', 'c'}: (3, 0, 1.72839), ...}
    dgraph = graph.copy()
    for edge in graph:
        e = tuple(edge)
        idxs = nodes.index(e[0]), nodes.index(e[1])
        edist = dists[*idxs]
        dgraph[edge] = tuple(graph[edge]) + (edist,)

    # iterate a constant number of times (should converge generally)
    for i in range(10):
        # while 1:

        # store euclidean distances and weights
        scores = {}
        for edge in graph:
            # get nodes and node indices
            node0, node1 = edge
            n0, n1 = nodes.index(node0), nodes.index(node1)
            # get distance and score
            edist = dists[n0, n1]
            dist = 0.
            for good_edge in range(graph[edge][0]):
                dist += edist
            for bad_edge in range(graph[edge][0]):
                dist += -alpha * edist
            #


        # calculate COM of all Nodes
        centers = {}
        for node in nodes:
            centers[node] = sum(-np.array(duv[node]) * np.array(wuv[node]))
            logger.info((node, centers[node]))

        # move a Node closer to that center of mass
        break


def plot_sphere(sphere, points, graph):
    """...
    """
    # ...
    e1 = []
    e2 = []
    w1 = []
    w2 = []
    nodes = sorted(frozenset.union(*graph))
    for edge, scores in graph.items():
        if scores[0]:
            e1.append(tuple(edge))
            w1.append(scores[0])
        if scores[1]:
            e2.append(tuple(edge))
            w2.append(scores[1])
    e1 = np.array(e1)
    e2 = np.array(e2)
    
    # ...
    c, a, m = toyplot.scatterplot(
        sphere[:, 0], sphere[:, 1],
        opacity=0.1,
        size=(sphere[:, 2] * 4) + 5,
        width=350,
        height=350,
    )
    a.graph(
        e1[:, 0],
        e1[:, 1],
        ewidth=w1,
        vcoordinates=points[:, :2],
        vlshow=False,
        vsize=0,
        #vsize=(points[:, 2] * 4) + 8,
        #vcolor='red',
        #vopacity=0.5,
        #vlstyle={"font-size": "15px"},
    )
    a.graph(
        e2[:, 0],
        e2[:, 1],
        ewidth=w2,
        estyle={"stroke-dasharray": "4,2"},
        vcoordinates=points[:, :2] + 0.02,
        vlshow=False,
        vsize=0,
    )
    a.scatterplot(0, 0, color='black')
    a.text(points[:, 0], points[:, 1], nodes, color="black", style={"font-size": 15})
    return c, a, m


# def get_center_of_mass(distance_graph):
#     """Return the center of mass of a graph where edge distances
#     are measured as the sum of Euclidean distances between all good
#     edges and subtacting the Euclidean distances -alpha of bad edges.

#     The center of mass (or centroid) can be thought of as a central
#     vertex or a weighted average of the positions of all vertices. If
#     the graph has weighted vertices, the center of mass may take those
#     weights into account.

#     1. Calculate function(graph, alpha)
#     2. optimize alpha
#     3. get center of mass
#     4. get outlier node
#     5. shift towards COM
#     6. repeat from 1 for N iterations until converged.
#     """
#     # for edge, (good_edges, bad_edges) in graph.items():



def binary_search_optimize_alpha(dgraph):
    """Return alpha minimizing ratio of +/- scores using binary search.

    """
    # TODO: special care for alpha += 4
    # Singleton cuts have an alpha of 2 so we increase alpha to find
    # non-singleton cuts until we hit an empty cut.../?
    # rho is the ratio of good to bad edges.
    # rho <= 4 * nqrts
    # e.g., for 10 quartets rho must be less than 40?
    # Proof: There are exactly 4jQj good edges in the graph.
    # Since there is no perfect cut, any cut contains at least a
    # single bad edge and we get rho=4|Q|.
    # If alpha > 4|Q| and a nonempty cut is returned, then G has a perfect cut.

    high = 1000
    low = 0.
    alpha = (high - low) / 2.

    # optimize score to zero
    while 1:
        # get the max value
        search = {
            i: dgraph[i][0] * dgraph[i][2]
            + (dgraph[i][1] * dgraph[i][2] * -alpha)
            for i in dgraph
        }
        # score = max(search.values())
        score = sum(search.values())
        # logger.debug(f"{alpha} {score}")#{search}")

        # if score has converged then stop.
        if high == low:
            break
        # if score is negative then decrease alpha
        elif score < 0:
            high = alpha
            alpha = (high - low) / 2.
        # if score is positive but still far from zero increase alpha
        elif score > 0.1:
            low = alpha
            alpha = (high + low) / 2.
        else:
            break
    return score, alpha


def get_random_3D_hyperplane() -> tuple[float, float, float]:
    """Return a random hyperplane in 3D space through the origin.
    """
    # Generate a random normal vector (a, b, c)
    normal_vector = np.random.randn(3)  # Generates a 3D vector with normal distribution

    # Normalize the vector to ensure it represents a direction
    normal_vector /= np.linalg.norm(normal_vector)
    return tuple(normal_vector)


def partition_points_by_3D_hyperplane(points, a, b, c, d) -> tuple[list, list]:
    """Partition points in 3D space based on their position relative to a hyperplane.

    Parameters
    ----------
    points: ndarray
        An array of shape (N, 3) where N is the number of points.
    a, b, c, d: float
        Coefficients of the hyperplane equation ax + by + cz + d = 0.
    """
    above_hyperplane = []
    below_hyperplane = []

    for point in points:
        x, y, z = point
        # Evaluate the hyperplane equation
        position = a * x + b * y + c * z + d

        if position > 0:
            above_hyperplane.append(point)
        elif position < 0:
            below_hyperplane.append(point)
        # Optionally handle points on the hyperplane (position == 0)
    return np.array(above_hyperplane), np.array(below_hyperplane)


def get_maxcut_heuristic(graph):
    """Return the cut that maximizes the parameterized qmc score.

    This function uses a heuristic approach to propose cuts given
    that the search space is too large to exhaustively search all
    possible cuts. A greedy algorithm is implemented that starts with
    an empty cut w/ a score of zero and a better cut is returned only
    if its score is greater. If no cut is found then a lower value of
    alpha is tested.
    """
    # ...
    score = 0
    nodes = frozenset.union(*graph)
    ntaxa = len(nodes)


def get_maxcut_qmc(graph: dict[tuple[frozenset,frozenset]: int]) -> tuple[frozenset, frozenset]:
    """Return the cut that maximizes the parameterized qmc score.

    This tests every possible cut that bipartitions the taxon set
    into two. Each cut can either satisfy, violate, or defer the
    relationship in each quartet. Depending on the result, each
    quartet contributes positive or negative values to the score,
    which are stored as [positive, negative] counts. A parameter
    alpha is optimized that is multiplied by the negative counts
    to find the cut that has a minimum non-negative ratio. This
    effectively penalizes singleton cuts to ensure cuts that divide
    the taxon set at least into min groups of 2.
    """
    nodes = frozenset.union(*graph)
    ntaxa = len(nodes)

    # visit every non-singleton cut (e.g., (0, 1), (0, 1, 2), ...)
    # shown as one side with all remaining taxa assumed on other side.
    scores = {}
    for size in range(2, ntaxa - 2):
        cuts = itertools.combinations(nodes, size)

        # iterate over cuts
        for cut in cuts:
            cut = frozenset(cut)
            ocut = nodes - cut
            score = 0
            alpha = 0

            for name in cut:
                for oname in ocut:
                    cut_edge = frozenset((name, oname))
                    # skip if edge is not in the quartet set, which can
                    # occur for subset quartet data sets.
                    if cut_edge in graph:
                        _score, _alpha = graph[cut_edge]
                        score += _score
                        alpha += _alpha
            scores[cut] = (score, alpha)

    # find alpha minimizing ratio of +/- scores using binary search
    high = 0.01
    low = -1000.
    alpha = (high + low) / 2.
    while 1:
        search = {i: scores[i][0] + scores[i][1] * alpha for i in scores}
        score = max(search.values())
        # logger.debug(f"{alpha} {score}")#{search}")
        if high == low:
            break
        elif score < 0:
            low = alpha
            alpha = (high + low) / 2.
        elif score > 0.1:
            high = alpha
            alpha = (high + low) / 2.
        else:
            break

    # get the max non-negative parameterized score
    max_non_negative_score = max(search[i] for i in search)
    # get list of all cuts w/ the max score (there are often ties)
    hits = [i for i in search.items() if i[1] == max_non_negative_score]
    # count N non-artificial taxa in each hit
    hits = [(i, get_part_size(i[0])) for i in hits]
    # choose a cut from among those with largest N
    ss = sorted(hits, key=lambda x: x[1])[-1][0][0]
    # return the selected cut and its inverse
    logger.debug(f"alpha={alpha:4f}; cut={sorted(ss)}")
    return ss, nodes - ss


##################################################################


def get_subset_quartets(quartets, bipartition):
    """Return two sets of quartets that include new deferred quartets
    created from a full set of quartets and a cut.

    The deferred set would replace ((a,b),(c,f)) with ((a,b),(c, x))
    if 'f' is divided from 'abc' by the cut. The name 'x' appends an
    int (e.g., 'x1') and is incremented each time this func is called.
    """
    # select one side of the cut
    cut = bipartition[0]

    # cut1 or cut2 quartets will be passed on while violated or
    # satisfied quartets are discarded.
    cut1_quartets = defaultdict(int)
    cut2_quartets = defaultdict(int)

    # iterate over quartets
    # e.g., qrt is (frozenset({'a', 'b'}), frozenset('c', 'd'))
    art = ArtificialTaxon()
    for qrt in quartets:
        a, b = qrt

        # partial a
        if a.issubset(cut):
            # unaffected: assign to 1
            if b.issubset(cut):
                cut1_quartets[qrt] += 1
                # print(a, b, '1')
            # satisfied: discard
            elif b.isdisjoint(cut):
                pass
                # print(a, b, 'x satisfy 1')
            # deferred: assign modified to 1
            else:
                tup = frozenset(b.intersection(cut).union({art}))
                qrt = (a, tup)
                cut1_quartets[qrt] += 1
                # print(a, b, f'1 defer - {qrt}')

        elif a.isdisjoint(cut):
            # satisfy: discard
            if b.issubset(cut):
                pass
                # print(a, b, 'x satisfy 2')
            # unaffected: assign to 2
            elif b.isdisjoint(cut):
                cut2_quartets[qrt] += 1
                # print(a, b, '2')
            # deferred: assign to 2
            else:
                tup = frozenset(b.difference(cut).union({art}))
                qrt = (a, tup)
                cut2_quartets[qrt] += 1
                # print(a, b, f'2 defep - {qrt}')

        # a conflicts with cut
        else:
            if b.issubset(cut):
                tup = frozenset(a.intersection(cut).union({art}))
                qrt = (b, tup)
                cut1_quartets[qrt] += 1
                # print(a, b, f'1 defeb - {qrt}')
            elif b.isdisjoint(cut):
                tup = frozenset(a.difference(cut).union({art}))
                qrt = (b, tup)
                cut2_quartets[qrt] += 1
                # print(a, b, f'2 defex - {qrt}')
            else:
                pass
                # print(a, b, 'x conflicts')
    parts = (bipartition[0].union({art}), bipartition[1].union({art}))
    qsets = (dict(cut1_quartets), dict(cut2_quartets))
    return parts, qsets


def get_part_size(object) -> int:
    """Return size of a partition not counting artificial taxa"""
    return sum(1 for i in object if not isinstance(i, ArtificialTaxon))


class ArtificialTaxon:
    count = 0
    def __init__(self):
        self.name = f"x{ArtificialTaxon.count}"
        ArtificialTaxon.count += 1

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


# NOTE:
# Here I implemented a normalization approach that gives high weight
# to splits that occur 100% of the time, but only in a subset of trees,
# due to having one or more taxa that are not present in all trees.
# This approach does not work well and leads to odd placement of the
# rare taxa.
#
# def get_weighted_quartet_graph_from_multitree(
#     trees: MultiTree | list[ToyTree],
#     normalize_by_ntrees: bool = False,
# ) -> dict[tuple[frozenset,frozenset],float]:
#     """Return {quartet: weight} from a MultiTree or sequence of trees.

#     The weight of each quartet is calculated as the proportion of trees
#     that contain that quartet, where this proportion can be scaled by
#     either the total number of trees, or if the trees contain different
#     tips, then by the number of trees that contain all of the tips in
#     the quartet.

#     Parameters
#     ----------
#     trees: MultiTree | Sequence[ToyTree]
#         An input set of one or more ToyTree objects.
#     normalize_by_ntrees: bool
#         If True then the weight of each quartet is calculated by its
#         number of occurrences in the trees divided by the total number
#         of trees, rather than the subset of trees that contain its tips.
#         If all trees have the same tips this has no effect.
#     """
#     # build weighted graph from quartets
#     # ...
#     tip_counts = {}
#     qrt_counts = {}

#     for tree in trees:
#         # store qrt
#         for qrt in tree.enum.iter_quartets(type=frozenset, sort=True):
#             if qrt not in qrt_counts:
#                 qrt_counts[qrt] = 1
#             else:
#                 qrt_counts[qrt] += 1

#         # store tips
#         tips = frozenset(tree.get_tip_labels())
#         if tips not in tip_counts:
#             tip_counts[tips] = 1
#         else:
#             tip_counts[tips] += 1

#     # get weights as the proportion of trees that included this qrt
#     # of the trees that included that set of taxa.
#     qrt_weights = {}
#     if normalize_by_ntrees:
#         for qrt in qrt_counts:
#             qrt_weights[qrt] = qrt_counts[qrt] / len(trees)
#     else:
#         for qrt in qrt_counts:
#             ntrees_with_qrt_tips = 0
#             for tips in tip_counts:
#                 qset = frozenset.union(*qrt)
#                 if qset.issubset(tips):
#                     ntrees_with_qrt_tips += tip_counts[tips]
#             qrt_weights[qrt] = qrt_counts[qrt] / ntrees_with_qrt_tips
#     return qrt_weights


##################################################################


def old_get_maxcut_quartet_supertree(
    trees: Union[MultiTree, list[ToyTree]],
    seed: Union[int, np.random.Generator, None] = None,
    normalize_by_ntrees: bool = False,
) -> ToyTree:
    """Return an unrooted supertree inferred from weighted quartet
    frequencies extracted from a set of input trees with the same or
    different tip sets.

    Parameters
    ----------
    trees: MultiTree | Sequence[ToyTree]
        An input set of one or more ToyTree objects.
    seed: int | np.random.Generator | None
        A seed for randomly choosing among equal scoring cuts.
    normalize_by_ntrees: bool
        If True then the weight of each quartet is calculated by its
        number of occurrences in the trees divided by the total number
        of trees, rather than the subset of trees that contain its tips.
        If all trees have the same tips this has no effect.

    Example
    -------
    >>> trees = toytree.mtree([
    >>>     "((a,b),(c,d));",
    >>>     "((a,b),(c,d));",
    >>>     "((a,b),(c,d,f));",
    >>> ])
    >>> tree = get_maxcut_quartet_supertree(trees)
    """
    raise SystemExit(0)

    # ...
    while partitions:
        # sample next partition
        p = partitions.popleft()
        print('part', p)

        # create a Node
        node = Node(name="".join(sorted(p)))
        print('node', node)

        # connect it to a parent
        if not nodes:
            treenode = node
        else:
            nodes[p]._add_child(node)

        # if partition is terminal
        if len(p) == 1:
            node.name = list(p)[0]
            continue

        # elif partition is a cherry or multi
        if len(p) == 2:
            for child in p:
                node._add_child(Node(name=child))
            print('nodes', nodes)
            continue

        # create a subgraph containing the remaining samples
        subgraph = {e: w for (e, w) in graph.items() if e.issubset(p)}
        print('sub', subgraph)

        # if no more edges to split a partition (e.g., hard polytomy)
        # if not subgraph:
        #     for child in p:
        #         node._add_child(Node(name=child))
        #     print('nodes', nodes)
        #     raise SystemExit()
        #     continue

        # get a maxcut bipartition and add its partitions to the queue
        bipart, score = get_maxcut(subgraph, seed=seed)

        # if no cut is made (hard polytomy)
        if not all(bipart):
            for child in p:
                node._add_child(Node(name=child))
            print('nodes', nodes)
            # raise SystemExit()
            continue

        print('bipart', bipart)
        for part in bipart:
            part = frozenset(part)
            partitions.append(part)
            nodes[part] = node

    # construct a ToyTree
    tree = toytree.ToyTree(treenode)

    # compute edge supports as the proportion of quartets supporting the split.
    for node, bipart in zip(tree[tree.ntips:], tree.enum.iter_bipartitions(type=frozenset)):
        matched = 0
        disjunct = 0
        b0, b1 = bipart
        for qrt in wqrts:
            q0, q1 = qrt
            if b0.issuperset(q0):
                if b1.issuperset(q1):
                    matched += wqrts[qrt]
                    print(q0, q1, b0, b1, wqrts[qrt], True)
            elif b0.issuperset(q1):
                if b1.issuperset(q0):
                    matched += wqrts[qrt]
                    print(q0, q1, b0, b1, wqrts[qrt], True)
            else:
                # this quartet is not relevant to the bipartition
                disjunct += wqrts[qrt]
        node.support = matched / (sum(wqrts.values()) - disjunct)
    return tree



def get_maxcut_quartet_supertree(
    trees: Union[MultiTree, list[ToyTree]],
    seed: Union[int, np.random.Generator, None] = None,
    use_weights: bool = False,
) -> ToyTree:
    """Return an unrooted supertree inferred from weighted quartet
    frequencies extracted from a set of input trees with the same or
    different tip sets.

    Parameters
    ----------
    trees: MultiTree | Sequence[ToyTree]
        An input set of one or more ToyTree objects with the same or
        different number of tips.
    seed: int | np.random.Generator | None
        A seed for randomly choosing among equal scoring cuts.
    use_weights: bool
        If True then quartet edges are weighted by their relative
        occurrence in the input trees.

    Example
    -------
    >>> trees = toytree.mtree([
    >>>     "((a,b),(c,d));",
    >>>     "((a,b),(c,d));",
    >>>     "((a,b),(c,d,f));",
    >>> ])
    >>> tree = get_maxcut_quartet_supertree(trees)
    """
    # get all quartets from the set of trees weighted by relative frequency
    wqrts = get_weighted_quartets_from_trees(trees)
    if not wqrts:
        raise ValueError("Cannot infer a quartet super tree for <4 taxa")
    if len(wqrts) == 1:
        raise ValueError("Cannot infer a quartet super tree for 4 taxa")
        return list(wqrts)

    # start iterative QMC
    queue = [wqrts]
    clades = []
    iteration = 0
    while queue:
        # get next quartet set to process
        wqrts = queue.pop()
        logger.info(("popped", wqrts))
        # build a graph of weighted edges from quartets
        graph = get_graph_from_quartets_qmc(wqrts)
        # find a maxcut
        bipart = get_maxcut_qmc(graph)
        logger.info((iteration, bipart))
        # add deferred quartet labels to cut and get subset quartets
        parts, qsets = get_subset_quartets(wqrts, bipart)
        logger.info((iteration, parts, qsets))
        # save finished partition to add quartet set to queue
        for part, qset in zip(parts, qsets):
            # resolved cherry
            if get_part_size(part) == 2:
                logger.warning(part)
                clades.append(part)
            # finished but unresolved polytomy
            elif not qset:
                logger.warning(part)
                clades.append(part)
            else:
                queue.append(qset)
        iteration += 1

    # build tree from clades
    # ... (clades)

    # assign supports from original quartet set
    # ... (tree, wqrts)
    return clades


def build_tree_from_deferred_clades(clades: list[frozenset]):
    """
    # build a tree from defereed clades, e.g.,
    # {a, b, x}, {c, d, y}, {e, f, x, y},               # (((a,b),(c,d)),(e,f));
    # {a, b, x}, {c, d, y}, {e, z, x, y}, {z, f, g, h}  # ((((a,b),(c,d)),e),(f,g,h));
    """
    nodes = {}
    for clade in clades:
        arts = frozenset({i for i in clade if isinstance(i, ArtificialTaxon)})
        children = clade - arts
        if arts in nodes:
            logger.warning(f"REPEAT {arts}")
        nodes[arts] = children
        # {{x}: (a, b), {y}: (c, d), {x,y}: (e,f)}
        # {{x}: (a, b), {y}: (c, d), {z}: (f,g,h), {x,y,z}: (e,)}
    logger.debug(('nodes', nodes))

    # iterate over artifical taxon set keys from smaller to larger
    snodes = sorted(nodes, key=len)
    tnodes = {}
    for aset in snodes:
        # create a parent Node and connect children
        if len(aset) == 1:
            parent = Node()
            for child in nodes[aset]:
                parent._add_child(Node(name=child))
            tnodes[aset] = parent
        # create a parent Node and connect children and other Nodes
        else:
            parent = Node()
            for child in nodes[aset]:
                parent._add_child(Node(name=child))
            for art in aset:
                anode = tnodes[frozenset({art})]
                parent._add_child(anode)
            tnodes[aset] = parent

        # for art in aset:
        #     parent._add_child(nodes[art])
    logger.debug(tnodes)
    tnodes[aset].draw_ascii()
    raise SystemExit(0)


if __name__ == "__main__":

    import toytree
    toytree.set_log_level("DEBUG")

    trees = [
        # "((a,b),(c,d));",
        # "((a,b),(c,d));",
        # "((a,b),(c,d));",
        # "((a,b),(c,d));",
        # "((a,b),(c,d));",
        # "((a,b),(c,d));",
        # "((a,b),(c,d));",
        # # "((b,a),((d,c),(g,f,e)));",
        # "((b,a),((d,c),(g,(f,e))));",  # problem child.
        # "((b,a),((d,c),((g,h),(f,e))));",  # problem child.
        # "((1,2),((3,4),(5,6)));",
        # "((1,2),(3,4));",
        # "((1,3),(4,5));",
        # "((3,4),(5,6));",
        # "((3,4),(5,7));",
        "((c,d),(a,b));",
        "((b,a),(c,d));",
        "((b,a),(d,c,f));",
        "((b,a),((d,c),f));",
        # "((b,a),((d,c),((f,e),g)));",
        # "((e,d),(h,a));",
        # "((a,c),((b,d),f));",
    ]
    logger.info(trees)
    # trees = [toytree.rtree.unittree(8, seed=123) for i in range(5)]
    trees = toytree.mtree(trees)
    # trees[-1].treenode.draw_ascii()
    # print(trees)
    clades = get_maxcut_quartet_supertree(trees)
    tree = build_tree_from_deferred_clades(clades)
    print(tree)
    raise SystemExit(0)

    # print('weighted quartets---------')
    q = get_weighted_quartets_from_trees(trees)
    # for i in q:
    #     print(i, q[i])

    # print('graph------------')
    g = get_graph_from_quartets_qmc(q)
    for i in g:
        print(i, g[i])

    # get the first cut
    # bipart = get_maxcut_qmc(g)
    # parts, qsets = get_subset_quartets(q, bipart)

    # # process each partition
    # for part, qset in zip(parts, qsets):
    #     nqrts = len(qset)
    #     print(part, qset)
    #     if get_part_size(part) == 2:
    #         logger.info(part)
    #     else:
    #         # for q in qset:
    #         #     print(q, qset[q])
    #         subg = get_graph_from_quartets_qmc(qset)
    #         bipart = get_maxcut_qmc(subg)
    #         parts, qsets = get_subset_quartets(qset, bipart)
    #         for part, qset in zip(parts, qsets):
    #             nqrts = len(qset)
    #             if get_part_size(part) == 2:
    #                 logger.info(part)
    #             else:
    #                 subg = get_graph_from_quartets_qmc(qset)
    #                 bipart = get_maxcut_qmc(subg)
    #                 parts, qsets = get_subset_quartets(qset, bipart)
    #                 if not nqrts:
    #                     logger.info(part)

            # for part, q in zip(bipart, qsets):
            #     nqrts = len(q)
            #     print(f"bipart {part} {nqrts} qrts ---------------")
            #     if nqrts > 2:
            #         subg = get_graph_from_quartets_qmc(q)
            #         bipart = get_maxcut_qmc(subg)
            #         qsets = get_subset_quartets(q, bipart[0])
            #         for part, q in zip(bipart, qsets):
            #             nqrts = len(q)
            #             print(f"bipart {part} {nqrts} qrts ---------------")



    # print("cut bipart---------------", q1)
    # subg = get_graph_from_quartets_qmc(q2)
    # print(subg)
    # bipart = get_maxcut_qmc(subg)
    # q1, q2 = get_subset_quartets(q1, bipart[1])
    # print(bipart)


    # subg = get_graph_from_quartets_qmc(q2)
    # bipart = get_maxcut_qmc(subg)
    # print(bipart)

    # subg = get_graph_from_quartets_qmc(q1)
    # bipart = get_maxcut_qmc(subg)
    # print(bipart)
    # q1, q2 = get_subset_quartets(q, bipart[1])
    # for i in q1:
    #     print(i, q1[i])
    # for i in q2:
    #     print(i, q2[i])




    # print('graph------------')
    # g = get_graph_from_quartets_qmc(q)
    # for i in g:
    #     print(i, g[i])



#!/usr/bin/env python

"""Outline of ideas for more phylocom metrics...

"""

from typing import TypeVar, Callable, Optional
from dataclasses import dataclass
from numpy.typing import ArrayLike
import numpy as np
import pandas as pd
import toytree


ToyTree = TypeVar("ToyTree")


@dataclass
class NullModel:
    tree: ToyTree
    """: Phylogenetic tree containing the complete species pool."""
    matrix: ArrayLike
    """: Community matrix shape=(nsites, nspecies), dtype=bool or int."""
    method: Callable
    """: A function to compute on the tree x matrix data."""
    size: int
    """: N replicate null community samples for calculating effect sizes."""
    _statistic: ArrayLike
    """: Statistic to be computed for each community."""
    _mean: float
    """: Mean statistic computed across null communities."""
    _std: float
    """: Standard deviation of statistic computed across null communities."""

    def __post_init__(self):
        """Sample null communities and return computed Z scores."""
        self._check_data()
        self._sample_from_null()
        self.get_standard_effect_size()

    def _check_data(self):
        """Verify that names match between tree and matrix.

        Names present in the matrix (DataFrame) should be present in
        the phylogeny, and vice-versa.
        """
        # TODO.

    def _sample_from_null(self):
        """Child class function will set self._mean and self._std."""
        raise NotImplementedError("No function, use a child class.")

    def get_standard_effect_size(self):
        r"""Return effect size (Z-value).
        
        $$ z = \frac{x - \mu}{\sigma} $$
        """
        return (self._statistic - self._mean) / self._std


class ShuffleTips(NullModel):
    """Shuffle tip labels on tree and recompute statistic."""
    def _sample_from_null(self):
        for rep in range(self.size):
            pass 

class Richness(NullModel):
    """Randomize community abundances while maintaining species richness"""

class Frequency(NullModel):
    """Randomize community abundances while maintaining species frequencies"""

class SamplePool(NullModel):
    """Randomize community by sampling species from pool of species 
    occurring in at least one community (sample pool) with equal 
    probability."""

class PhyloPool(NullModel):
    """Randomize community by sampling species from pool of species 
    occurring in at least one community (sample pool) with equal 
    probability.
    
    Note: this description is same as above, probably a typo.
    """

class IndependentSwap(NullModel):
    """Randomizes community data matrix with the independent swap 
    algorithm (Gotelli 2000) maintaining species occurrence frequency
    and sample species richness."""

class TrialSwap(NullModel):
    """Randomizes community data matrix with the trial-swap algorithm
    (Miklos & Podani 2004) maintaining species occurrence frequency
    and sample species richness."""



def get_faith_pd() -> float:
    """..."""


def get_mean_pairwise_distance() -> float:
    """..."""


def get_mean_nearest_neighbor_distance() -> float:
    """..."""


def get_net_relatedness_index() -> float:
    """..."""


def get_nearest_taxon_index() -> float:
    """..."""


def get_species_richness() -> float:
    """..."""


def get_phylogenetic_sorensen_index() -> float:
    """Return the Phylogenetic Sørensen Index of PhyloBetaDiversity.

    This is a phylogenetic analog of the Sørensen index based on 
    the total length of edges shared and unshared between paired
    communities.

    References
    ----------
    - Bryant et al., 2008; Swenson, 2011; Feng et.al., 2012)

    Examples
    --------
    >>> import toyplot, toytree
    >>> tree = ...
    >>> matrix = ...
    >>> sor_index = get_phylo...
    >>>
    >>> # plot a matrix of SorIndex values between pairs of comms.
    >>> ...
    """

def get_unifrac() -> float:
    """Return the Unifrac value of PhyloBetaDiversity.
    
    This measures the unique fraction of the phylogeny that is 
    unshared between a pair of communities. This metric is mostly
    sensitive to dissimilarity at the tips of a phylogeny, but not
    among splits farther back in a tree.
    """



#!/usr/bin/env python

"""Trait evolution subpackage.

"""

# from toytree.pcm.src.traits.discrete_markov_model_sim import (
#     get_markov_model,
#     simulate_discrete_data,
# )

# from toytree.pcm.src.traits.continuous_brownian_sim import (
#     simulate_continuous_brownian,
# )



#!/usr/bin/env python

"""Add Pie Chart Marks at nodes or edges.

"""

from typing import Sequence, Union, TypeVar

from loguru import logger
import numpy as np
from toytree.core import ToyTree, Cartesian, Mark
from toytree.style import check_arr, get_color_mapped_values
from toytree.annotate.src.checks import get_last_toytree_mark, assert_tree_matches_mark
from toytree.core.apis import add_subpackage_method, AnnotationAPI
from toytree.annotate.src.add_edge_markers import get_edge_midpoints
from toytree.drawing.src.mark_pie import PieChartMark
from toytree.style.src.validate_data import (
    validate_numeric,
    validate_mask,
)

logger = logger.bind(name="toytree")
Color = TypeVar("Color", str, tuple, np.ndarray)
__all__ = [
    "add_node_pie_charts",
    "add_edge_pie_charts",
]


def validate_pie_data(
    tree: ToyTree,
    data: np.ndarray,
    # min_size: float = 1e-9,
    # mask: np.ndarray = None,
) -> np.ndarray:
    """Return cleaned pie chart data.

    Checks pie chart data for correct shape and type, and that the row
    values sum to 1.

    Formats
    -------
    - 1-D array -> 2-D array
    - 2-D array -> 2-D array sums to 1
    """
    assert data.min() >= 0, "negative values are not allowed in pie chart data."
    assert data.shape[0] in (tree.nnodes, tree.nnodes - 1), (
        f"pie chart data must be shape (nnodes, nvalues), your data is {data.shape}.")

    # allow single value arrays in [0, 1] to represent two categories
    if data.ndim == 1:
        if data.max() > 1:
            raise ValueError(
                "1 dimensional array data for pie charts must be < 1 to "
                "be expanded to 2 categories: (value, 1 - value).")
        return np.column_stack([data, 1 - data])

    # 2D arrays represent (ntips, ntraits) data.
    else:
        assert np.allclose(data.sum(axis=1), 1), "pie chart data row values must sum to 1."
    return data


@add_subpackage_method(AnnotationAPI)
def add_node_pie_charts(
    tree: ToyTree,
    axes: Cartesian,
    data: np.ndarray,
    size: Union[int, Sequence[int]] = 10,
    colors: Union[Sequence[Color], Color] = None,
    ostroke: Color = "#262626",
    ostroke_width: float = 1.5,
    istroke: Color = "#262626",
    istroke_width: float = 0.,
    rotate: int = -45,
    mask: Union[bool, np.ndarray, tuple] = False,
    xshift: int = 0,
    yshift: int = 0,
) -> Mark:
    """Return a toyplot Mark of node markers added to a tree plot.

    This adds node markers to the last tree drawn on the Cartesian
    axes. The shape, size, color, and style of markers can be modified.

    Parameters
    ----------
    axes: Cartesian
        A toyplot Cartesian axes object containing a tree drawing.
    data: numpy.ndarray
        Array of shape(ncategories, nnodes) with rows summing to 1. 
    size: int or Sequence[int]
        Size of markers as single int or Sequence of ints, in px units.
    colors: None, str, tuple, or array, or Sequence
        Color for each category/trait or the name of a colormap.
    ostroke: Color
        Color of the stroke on the outside of the Mark.
    ostroke_width: float
        Width of the stroke on the outside of the Mark
    istroke: Color
        Color of the stroke on the inside of the Mark between wedges.
    istroke_width: float
        Width of the stroke on the inside of the Mark between wedges.
    rotate: int
        Rotate the starting point of the wedges.
    mask: bool, np.ndarray, or tuple
        Node mask to hide/show some or all Nodes.
    xshift: int
        Shift marker horizontally by px units (+=right, -=left).
    yshift: int
        Shift marker vertically by px units (+=down, -=up).

    Example
    -------
    >>> tree = toytree.rtree.unittree(6, seed=123)
    >>> canvas, axes, m0 = tree.draw()

    >>> # generate random pie-like (proportion) data array
    >>> import numpy as np
    >>> ncategories = 3
    >>> arr = np.random.random(size=(tree.nnodes, ncategories))
    >>> arr = (arr.T / arr.sum(axis=1)).T

    >>> # add pie charts to all internal Nodes
    >>> tree.annotate.add_node_pie_charts(
    >>>     axes=axes, data=arr, size=20, mask=(0, 1, 1),
    >>>     istroke_width=0.75, istroke="black", rotate=-45,
    >>> )
    """
    # get mark for coordinates on plotted tree.
    mark = get_last_toytree_mark(axes)
    assert_tree_matches_mark(tree, mark)

    # get mask
    mask = validate_mask(tree, style={"node_mask": mask})

    # check and cleanup data input.
    # TODO: option to collapse categories below a minimum percentage
    data = validate_pie_data(tree, data)

    # expand colormap to an array of colors
    if colors is None:
        colors = "Set2"
    if isinstance(colors, (tuple, list, np.ndarray)):
        pass
    else:
        colors = get_color_mapped_values(range(data.shape[1]), colors)

    # ensure conversion of colors to array type and size=ncategories
    colors = check_arr(
        values=colors,
        label="colors (pie chart category colors)",
        size=data.shape[1],
        ctype=np.void,
    )
    sizes = validate_numeric(
        tree, key="size", size=tree.nnodes, style={"size": size})[mask]

    # mask some Nodes
    data = data[mask, :]
    coords = mark.ntable[mask, :]

    # plot edge markers as scatterplot markers
    mark = PieChartMark(
        coordinates=coords,
        data=data,
        sizes=sizes,
        colors=colors,
        ostroke=ostroke,
        ostroke_width=ostroke_width,
        istroke=istroke,
        istroke_width=istroke_width,
        rotate=rotate,
        xshift=xshift,
        yshift=yshift,
    )
    axes.add_mark(mark)
    return mark


@add_subpackage_method(AnnotationAPI)
def add_edge_pie_charts(
    tree: ToyTree,
    axes: Cartesian,
    data: np.ndarray,
    size: Union[int, Sequence[int]] = 10,
    colors: Union[Sequence[Color], Color] = None,
    ostroke: Color = "#262626",
    ostroke_width: float = 1.5,
    istroke: Color = "#262626",
    istroke_width: float = 0.,
    rotate: int = -45,
    mask: Union[bool, np.ndarray, tuple] = False,
    xshift: int = 0,
    yshift: int = 0,
) -> Mark:
    """Return a toyplot Mark of edge pie charts added to a tree plot.

    This adds edge pie chart markers to the last tree drawn on the
    Cartesian axes.

    Parameters
    ----------
    axes: Cartesian
        A toyplot Cartesian axes object containing a tree drawing.
    data: numpy.ndarray
        Array of shape(ncategories, nnodes) with rows summing to 1. 
    size: int or Sequence[int]
        Size of markers as single int or Sequence of ints, in px units.
    colors: None, str, tuple, or array, or Sequence
        Color for each category/trait or the name of a colormap.
    ostroke: Color
        Color of the stroke on the outside of the Mark.
    ostroke_width: float
        Width of the stroke on the outside of the Mark
    istroke: Color
        Color of the stroke on the inside of the Mark between wedges.
    istroke_width: float
        Width of the stroke on the inside of the Mark between wedges.
    rotate: int
        Rotate the starting point of the wedges.
    mask: bool, np.ndarray, or tuple
        Node mask to hide/show some or all Nodes.
    xshift: int
        Shift marker horizontally by px units (+=right, -=left).
    yshift: int
        Shift marker vertically by px units (+=down, -=up).

    Example
    -------
    >>> tree = toytree.rtree.unittree(6, seed=123)
    >>> canvas, axes, m0 = tree.draw()

    >>> # generate random pie-like (proportion) data array
    >>> import numpy as np
    >>> ncategories = 3
    >>> arr = np.random.random(size=(tree.nnodes, ncategories))
    >>> arr = (arr.T / arr.sum(axis=1)).T

    >>> # add pie charts to all internal Nodes
    >>> tree.annotate.add_edge_pie_charts(
    >>>     axes=axes, data=arr, size=20, mask=False,
    >>>     istroke_width=0.75, istroke="black", rotate=-45,
    >>> )
    """
    # get mark for coordinates on plotted tree.
    mark = get_last_toytree_mark(axes)
    assert_tree_matches_mark(tree, mark)

    # get coordinates of all real edges
    nedges = tree.nnodes - 2 if tree.is_rooted() else tree.nnodes - 1
    #coords = _get_edge_midpoints(tree, mark.ntable, mark.layout, mark.edge_type)
    coords = get_edge_midpoints(mark.etable, mark.ntable, mark.layout, mark.edge_type)

    mask = validate_mask(tree, style={"node_mask": mask})[:nedges]

    # validate data.
    # TODO: option to collapse categories below a minimum percentage
    data = validate_pie_data(tree, data)

    # expand colormap to ncolor (note: not nedges)
    if colors is None:
        colors = "Set2"
    if isinstance(colors, (tuple, list, np.ndarray)):
        pass
    else:
        colors = get_color_mapped_values(range(data.shape[1]), colors)

    # ensure conversion of colors to array type and size=ncategories
    colors = check_arr(
        values=colors,
        label="colors (pie chart category colors)",
        size=data.shape[1],
        ctype=np.void,
    )

    # mask some Nodes
    data = data[:nedges][mask, :]
    sizes = validate_numeric(
        tree, key="size", size=tree.nnodes, style={"size": size})[:nedges][mask]

    coords = coords[:nedges][mask, :]

    # plot edge markers as scatterplot markers
    mark = PieChartMark(
        coordinates=coords,
        data=data,
        sizes=sizes,
        colors=colors,
        ostroke=ostroke,
        ostroke_width=ostroke_width,
        istroke=istroke,
        istroke_width=istroke_width,
        rotate=rotate,
        xshift=xshift,
        yshift=yshift,
    )
    axes.add_mark(mark)
    return mark


if __name__ == "__main__":

    import toytree
    tree = toytree.rtree.unittree(6, seed=123)
    canvas, axes, m0 = tree.draw()
    # generate random pie-like (proportion) data array
    import numpy as np
    ncategories = 3
    arr = np.random.random(size=(tree.nnodes, ncategories))
    arr = (arr.T / arr.sum(axis=1)).T

    # add pie charts to all internal Nodes
    tree.annotate.add_node_pie_charts(
        axes=axes,
        data=arr,
        size=20,
        mask=False,
        istroke_width=0.75,
        istroke="black",
        rotate=0,
        colors="Greys"
    )
    toytree.utils.show(canvas)



#!/usr/bin/env python

"""Save canvas to file rendered as HTML, SVG, PDF, or PNG.

This is similar to the toyplot.render module but is intended to be a
bit simpler by requiring import of the file format prior to use. By
creating a separate module for this in toytree it also reduces the
need to import toyplot alongside toytree solely for the purpose of
saving canvases.
"""

from typing import Union
from pathlib import Path
from toytree.core import Canvas


SUFFIXES = (".html", ".svg", ".pdf", ".png")


def save(canvas: Canvas, path: Union[str, Path]):
    """Save a Canvas to a file path.

    A Canvas can be saved in a variety of formats. If no recognized
    filename suffix exists in the path argument then the default format
    of HTML will be used. Other options are SVG, PDF, and PNG.

    Parameters
    ----------
    canvas: toyplot.Canvas
        A Canvas object created from a toyplot or toytree drawing.
    path: str | Path
        A filepath at which to save the canvas. The suffix of the path
        is used to designate the file type. Default is HTML, but SVG,
        PDF, and PNG are also supported.

    Example
    -------
    tree = toytree.rtree.coaltree(10)
    canvas, axes, mark = tree.draw(ts='c')
    toytree.save(canvas, "./drawing.svg")
    """

    # get suffix for path
    path = Path(path)
    suffix = path.suffix

    # if not suffix (only filename) then use html.
    if not suffix:
        suffix = ".html"

    # if suffix not recognized then raise error.
    if suffix.lower() not in SUFFIXES:
        raise OSError(
            "File path suffix not recognized ({suffix}). "
            "Path should end with one of {SUFFIXES}.")

    # ...
    if suffix == ".html":
        import toyplot.html
        toyplot.html.render(canvas, str(path.with_suffix(suffix)))

    elif suffix == ".svg":
        import toyplot.svg
        toyplot.svg.render(canvas, str(path.with_suffix(suffix)))

    elif suffix == ".pdf":
        import toyplot.pdf
        toyplot.pdf.render(canvas, str(path.with_suffix(suffix)))

    elif suffix == ".png":
        import toyplot.png
        # TODO: accommodate pip installs w/o ghostscript here
        toyplot.png.render(canvas, str(path.with_suffix(suffix)))


if __name__ == "__main__":
    pass



#!/usr/bin/env python

"""Split trees into bipartitions and return in a variety of formats.

Methods
-------
iter_bipartitions

"""

from typing import TypeVar, Iterator, Tuple, Optional, Set, Callable, Sequence
from loguru import logger
from toytree import Node, ToyTree
from toytree.core.apis import TreeEnumAPI, add_subpackage_method, add_toytree_method

logger = logger.bind(name="toytree")
Query = TypeVar("Query")

__all__ = [
    "iter_bipartitions",
    "_iter_bipartition_sets",
]


@add_subpackage_method(TreeEnumAPI)
def _iter_bipartition_sets(
    tree: ToyTree,
    feature: Optional[str] = "name",
    include_singleton_partitions: bool = False,
    include_internal_nodes: bool = False,
    normalize: bool = False,
) -> Iterator[Tuple[Set[Node], Set[Node]]]:
    """Generator to yield bipartitions as tuples of two sets of Nodes.

    Bipartitions are yielded in idxorder as Tuple[set, set] objects
    where each represents nodes on either side of an edge. Each is
    ordered as (child, parent) pairs such that the first set contains
    Nodes below the edge and the second Nodes above the edge.

    The order in which bipartitions are yielded, as well as the order
    of the two partitions within each bipartition, depends on the tree
    topology and rooting. To get collections of bipartitions that are
    the same for a given topology regardless of its rooting see the
    options in `.iter_bipartitions`.

    Parameters
    ----------
    feature: str or None
        Option to represent Nodes in bipartitions by a feature, such
        as 'name' or 'idx'. If None then Node objects are returned.
    include_singleton_partitions: bool
        If True then singleton splits (e.g., (A | B,C,D)) are included
        in the result. By default these are excluded since it is
        implicit that one exists for every tip Node in a tree.
    include_internal_nodes: bool
        Default is to only show tip Nodes on either side of a
        bipartition, but internal Nodes can be included as well. In
        this case the results are easier to interpret if internal Nodes
        have names assigned, else you can set the 'feature' arg to
        None, or 'idx', to return Node objects, or a unique feature.

    Examples
    --------
    >>> newick = "(((a,b)X,((c,d)Y,e)Z)R;"
    >>> tree = toytree.tree(newick)

    >>> list(tree.enum._iter_bipartition_sets("name""))
    >>> # [({'c', 'd'}, {'b', 'e', 'f', 'z'}),
    >>> #  ({'e', 'f'}, {'b', 'c', 'd', 'z'}),
    >>> #  ({'c', 'd', 'e', 'f'}, {'b', 'z'})]

    >>> list(tree.enum._iter_bipartition_sets("idx", include_internal_nodes=True))
    >>> # [({2, 3, 6}, {0, 1, 4, 5, 7, 8, 9}),
    >>> #  ({4, 5, 7}, {0, 1, 2, 3, 6, 8, 9}),
    >>> #  ({2, 3, 4, 5, 6, 7, 8}, {0, 1, 9})]
    """
    # store cache of desc below each node to reduce traversals
    cache = {}

    # nodes set to iterate over. Do not include the root.
    node_set = set(tree)
    topnode = tree.nnodes - 1

    # do not include root node in node_set if tree is rooted.
    if tree.is_rooted():
        node_set -= {tree.treenode}
        topnode -= 1

    # iterate over all nodes in idx order building cache as it goes.
    for node in tree[:topnode]:
        if node.is_leaf():
            below = cache[node] = {node}
            if not include_singleton_partitions:
                continue
        else:
            below = set.union(*([{node}] + [cache[i] for i in node.children]))
            cache[node] = below

        # get all nodes not under this split
        other = node_set - cache[node]

        # if excluding internal Nodes
        if not include_internal_nodes:
            below = set(i for i in below if i._idx < tree.ntips)
            other = set(i for i in other if i._idx < tree.ntips)

        # yield biparts except at root in unrooted trees: ({}, {all})
        if feature is None:
            yield below, other
        else:
            yield (
                set(getattr(i, feature) for i in below),
                set(getattr(i, feature) for i in other),
            )


@add_toytree_method(ToyTree)
@add_subpackage_method(TreeEnumAPI)
def iter_bipartitions(
    tree: ToyTree,
    feature: Optional[str] = "name",
    include_singleton_partitions: bool = False,
    include_internal_nodes: bool = False,
    type: Callable = set,
    sort: bool = False,
) -> Iterator[Tuple[Sequence, Sequence]]:
    """Generator of bipartitions (Nodes on either side of edges).

    Bipartitions represent the splits in a tree. Many algorithms compare
    tips (or internal Nodes) on either side of each split to compute
    metrics on trees. This function aims to provide a flexible and fast
    framework for yielding bipartitions in various formats.

    Notes
    -----
    - Bipartitions are generated in Node idx traversal order.
    - Bipartitions are formatted as a tuple of two items, each of
    which is referred to as a partition.
    - The order of partitions, e.g. (part1, part2) can be toggled using
    the argument `sort`.
    - The type used to represent a partition can be toggled using the
    argument `type`. Common formats are `set` or `tuple`.

    Parameters
    ----------
    feature: str
        Feature to return to represent Nodes on either side of a
        bipartition. Default is "name". None will return Node objects.
        Any other Node feature, such as "idx" can also be used. Note
        the feature arg does not affect the order in which partitions
        or bipartitions are returned/sorted (see `sort` argument below).
    include_singleton_partitions: bool
        If True then singleton splits (e.g., (A | B,C,D)) are included
        in the result. By default these are excluded since it is
        implicit that one exists for every tip Node in a tree.
    include_internal_nodes: bool
        Default is to only show tip Nodes on either side of a
        bipartition, but internal Nodes can be included as well. In
        this case the results are easier to interpret if the returned
        values have unique features (e.g., feature=None or 'idx').
    type: Callable
        The type of collection used to represent a partition. Default
        is `set` to return a tuple of sets, but another useful option
        is `tuple`, which returns a tuple of tuples. The latter
        collection can be converted into a set of bipartitions.
    sort: bool
        If False, bipartitions are returned as (child, parent) order
        given the topology and rooting in Node idx order traversal. If
        sort=True, bipartitions are instead always sorted first by len,
        e.g., (fewer, longer) and if the same len, then next by the
        lowest alphanumeric tip name, e.g., ({'a', 'b'}, {'c', 'd'}).
        If the requested partition `type` is sortable (i.e., not a set)
        then items within a partition are also consistently sorted.

    Examples
    --------
    >>> tree = toytree.tree("(a,b,((c,d)CD,(e,f)EF)X)AB;")

    >>> # default: biparts w/ tip names as (child, parent) in idx order
    >>> list(tree.iter_bipartitions())
    >>> # [({'c', 'd'}, {'a', 'b', 'e', 'f'}),
    >>> #  ({'e', 'f'}, {'a', 'b', 'c', 'd'}),
    >>> #  ({'c', 'd', 'e', 'f'}, {'a', 'b'})]

    >>> # same order, but as int idx labels including all Nodes
    >>> list(tree.iter_bipartitions(
    >>>     feature='idx',
    >>>     include_internal_nodes=True,
    >>>     include_singleton_partitions=True,
    >>> ))
    >>> # [({0}, {1, 2, 3, 4, 5, 6, 7, 8, 9}),
    >>> #  ({1}, {0, 2, 3, 4, 5, 6, 7, 8, 9}),
    >>> #  ({2}, {0, 1, 3, 4, 5, 6, 7, 8, 9}),
    >>> #  ({3}, {0, 1, 2, 4, 5, 6, 7, 8, 9}),
    >>> #  ({4}, {0, 1, 2, 3, 5, 6, 7, 8, 9}),
    >>> #  ({5}, {0, 1, 2, 3, 4, 6, 7, 8, 9}),
    >>> #  ({2, 3, 6}, {0, 1, 4, 5, 7, 8, 9}),
    >>> #  ({4, 5, 7}, {0, 1, 2, 3, 6, 8, 9}),
    >>> #  ({2, 3, 4, 5, 6, 7, 8}, {0, 1, 9})]

    >>> # args to get consistently sorted biparts regardless of rooting
    >>> sorted(tree.iter_bipartitions(type=tuple, sort=True))
    >>> # [(('a', 'b'), ('c', 'd', 'e', 'f')),
    >>> #  (('c', 'd'), ('a', 'b', 'e', 'f')),
    >>> #  (('e', 'f'), ('a', 'b', 'c', 'd'))]

    >>> # convert consistent bipartitions to sets for easy comparison
    >>> x = set(tree.root('a').iter_bipartitions(type=tuple, sort=True))
    >>> y = set(tree.root('e').iter_bipartitions(type=tuple, sort=True))
    >>> assert x == y
    """
    kwargs = dict(
        tree=tree,
        feature=feature,
        include_singleton_partitions=include_singleton_partitions,
        include_internal_nodes=include_internal_nodes,
    )

    # fastest approach returns bipart consistently as (child, parent)
    # and get feature from _iter_bipartition_sets if requested.
    # yield ({part1}, {part2})
    if (not sort) and (type == set):  # noqa E721
        for below, other in _iter_bipartition_sets(**kwargs):
            yield below, other

    # slower approach sorts bipart alphanumerically by tip Node names.
    # sort requires first getting objects as Nodes so we can access
    # the Name feature regarldess of what the returned feature will be.
    else:
        # yield as type(*Nodes) or as type(feature(*Nodes))
        if feature is not None:
            def tformat(node_tuple):
                return type(getattr(i, feature) for i in node_tuple)
        else:
            tformat = type

        # iterate over bipartitions and order items within
        kwargs['feature'] = None
        for below, other in _iter_bipartition_sets(**kwargs):
            # optional: sort biparts by len or min name
            # always: if not type==set sort within by name
            # returned as sets or sorted lists
            clade1, clade2 = _format_bipartition(below, other, type, sort)

            # return as requested feature and type
            yield tformat(clade1), tformat(clade2)


def _build_node_names_for_sorting(node: Node) -> str:
    """Returns node name to use while sorting tip and internal nodes."""
    if node.is_leaf():
        return node.name
    return "".join(sorted(node.get_leaf_names())[::-1])


def _format_bipartition(
    below: Set,
    other: Set,
    type: Callable,
    sort: bool,
) -> Tuple[Sequence[Node], Sequence[Node]]:
    """Sort bipartitions ({Node, Node}, {Node, Node}).

    If sort is True this first sorts the two sides of the split by
    length, or by lowest alphanumeric name if they are same length.
    Example: ({Node, Node}, {Node, Node, Node})
    Example: ({'a', 'z'}, {'b', 'x'})
    This also requires assigning names to internal nodes when present,
    based on their descendant tip names.
    Example: ({'a', 'z', 'za'}, {'b', 'x', 'xb'})
    If the type is not set then items within partitions are
    consistently sorted.
    Example: (['a', 'z', 'za'], ['b', 'x', 'xb']) and can
    be converted to final type (e.g., tuple) in later steps.
    """
    blen = len(below)
    olen = len(other)

    # if type == set then items within a partition do not need to be
    # sorted, but part1 vs part2 does still need it.
    if sort and (type == set):  # noqa E721
        if blen < olen:
            return below, other
        elif olen < blen:
            return other, below
        else:
            # min names
            bmin = min(i.name for i in below)
            omin = min(i.name for i in other)
            if bmin < omin:
                return below, other
            else:
                return other, below

    # if sort == False and type != set then items within partition need
    # to be sorted, but part1 vs part2 does not b/c we want to maintain
    # the (child, parent) order.
    if not sort:
        return (
            sorted(below, key=_build_node_names_for_sorting),
            sorted(other, key=_build_node_names_for_sorting),
        )

    # if sort == True and type != set then call sort on each partition
    # as well as between the two partitions.
    if blen < olen:
        return (
            sorted(below, key=_build_node_names_for_sorting),
            sorted(other, key=_build_node_names_for_sorting),
        )
    elif olen < blen:
        return (
            sorted(other, key=_build_node_names_for_sorting),
            sorted(below, key=_build_node_names_for_sorting),
        )
    else:
        # min names
        bmin = min(_build_node_names_for_sorting(i) for i in below)
        omin = min(_build_node_names_for_sorting(i) for i in other)
        if bmin < omin:
            return (
                sorted(below, key=_build_node_names_for_sorting),
                sorted(other, key=_build_node_names_for_sorting),
            )
        else:
            return (
                sorted(other, key=_build_node_names_for_sorting),
                sorted(below, key=_build_node_names_for_sorting),
            )


if __name__ == "__main__":

    import toytree

    tree = toytree.rtree.unittree(10, seed=123)

    for bipart in tree.iter_bipartitions(
        'idx', sort=True, include_singleton_partitions=True):
        print(bipart)



#!/usr/bin/env python

"""Simulate one or more continuous traits under Brownian motion.

"""

from typing import Optional, Union, Sequence, Dict
import pandas as pd
import numpy as np
from toytree import ToyTree
from toytree.utils import ToytreeError
from toytree.core.apis import add_subpackage_method, PhyloCompAPI

__all__ = [
    "simulate_continuous_brownian",
]


def simulate_continuous_multivariate_brownian(
    tree: ToyTree,
    rates: Union[float, Sequence[float], Dict[str, float]],
    root_states: Union[float, Sequence[float], Dict[str, float], None] = None,
    seed: Union[int, np.random.Generator, None] = None,
    df: bool = False,
) -> Union[ToyTree, pd.DataFrame]:
    """Simulate two or more continuous correlated traits under Brownian motion.

    References
    ----------
    - https://lukejharmon.github.io/pcm/chapter3_bmintro/
    """
    # TODO


@add_subpackage_method(PhyloCompAPI)
def simulate_continuous_brownian(
    tree: ToyTree,
    rates: Union[float, Sequence[float]],
    root_states: Union[float, Sequence[float], None] = None,
    tips_only: bool = False,
    inplace: bool = False,
    seed: Union[int, np.random.Generator, None] = None,
) -> Union[pd.DataFrame, None]:
    """Simulate one or more continuous traits under Brownian motion.

    The amount of change in a continuous trait over a given time
    interval can be modeled under Brownian motion as the result of
    a random walk. At each time step the value changes by an amount
    randomly sampled from a normal distribution with mean=0 and variance
    described by an evolutionary rate parameter, sigma**2. To model the
    change over an interval of time we can simply sample a random value
    from a normal distribution with mean=0 and variance as the product
    of the length of time and the rate parameter sigma**2.

    Simulated traits are labeled t0-tN for N traits, unless the rates
    arg is entered as a mapping (e.g., dict) in which case traits can
    be given custom names. By default, simulated data are stored to
    Node objects of the input tree, and can be fetched by calling
    `tree.get_node_data()`. However, you can alternatively use the
    argument `df=True` to instead return simulated data in a DataFrame.

    Paramaters
    ----------
    tree: ToyTree
        A ToyTree on which traits will be simulated given the topology
        and branch lengths.
    rates: float, Sequence[float], or Mapping[str, float]
        Evolutionary rate parameter(s) (sigma**2) for one or more traits
        which determine the rate of the random walk.
    root_states: None, float, Sequence[float], or Mapping[str, float]
        The root state for each trait value. This is the expected mean
        state for all values at the tips under Brownian motion. If None
        (default) the root states are set to 0.
    tips_only: bool
        If True values are only returned for the tip Nodes.
    inplace: bool
        If False simulated data are returned as a pandas DataFrame; if
        True simulated data are assigned to Nodes on the input tree,
        and can be fetched using `tree.get_node_data()`.
    seed: int, numpy RNG, or None
        Seed for numpy random number generator.

    References
    ----------
    - https://lukejharmon.github.io/pcm/chapter3_bmintro/

    Example
    -------
    >>> tree = toytree.rtree.unittree(10, treeheight=10)
    >>> data = tree.pcm.simulate_continuous_brownian(rate=1.0)

    >>> # simulate multiple traits with different rates
    >>> data = tree.pcm.simulate_continuous_brownian(rates=[1, 2, 3])

    >>> # simulate 100 replicates w/ same rate and assign to tree Nodes
    >>> tree.pcm.simulate_continuous_brownian([1.0] * 100, inplace=True)
    >>> print(tree.get_node_data())
    """
    # seed random number generator
    rng = np.random.default_rng(seed)

    # make ratess into a dict w/ trait names (t0, t1, t2, ...) or custom
    if isinstance(rates, (float, int)):
        rates = np.array([rates])
        names = ["t0"]
    elif isinstance(rates, (list, tuple, np.ndarray)):
        rates = np.array(rates)
        names = [f"t{i}" for i in range(len(rates))]
    else:
        raise ToytreeError("rate argument type not supported")

    # check root states arg
    if root_states is None:
        root_states = np.zeros(len(rates))
    elif isinstance(root_states, (float, int)):
        root_states = np.array([root_states])
    elif isinstance(root_states, (list, tuple, np.ndarray)):
        root_states = np.array(root_states)
    else:
        raise ToytreeError("root_states argument type not supported.")
    if len(root_states) != len(rates):
        raise ToytreeError(
            "root_states and rates args must be same length, or root_states=None.")

    # create tree copy
    # tree = tree if inplace else tree.copy()

    # store trait values in array
    arr = np.zeros((tree.nnodes, len(root_states)))
    arr[-1] = root_states
    for node in tree[::-1][1:]:
        arr[node.idx] = rng.normal(arr[node.up.idx], node.dist * rates)

    # return as tree (default), array, or dataframe
    end = tree.ntips if tips_only else tree.nnodes
    if inplace:
        for node in tree[:end]:
            for tidx in range(arr.shape[1]):
                trait = names[tidx]
                setattr(node, trait, arr[node._idx, tidx])
        return None
    return pd.DataFrame(arr[:end], columns=names)


if __name__ == "__main__":

    import toytree
    tree = toytree.rtree.unittree(10, treeheight=1)
    data = simulate_continuous_brownian(
        tree,
        rates=[1., 2.],
        root_states=[10, 20],
        seed=123,
        tips_only=True,
    )
    print(data)

    data = simulate_continuous_brownian(
        tree,
        rates=1.,
        seed=123,
        tips_only=True,
    )
    print(data)

    # simulate_continous_brownian(tree, rates=1, seed=123, inplace=True)
    # tree._draw_browser(
    #     # edge_widths=4,
    #     # ts='c',
    #     # edge_style={"stroke-width": 4},
    #     node_sizes="t0",
    #     node_colors="t0",
    #     node_mask=False,
    #     node_labels=("t0", "{:.2f}"),
    #     node_labels_style={"anchor_shift": -10, "baseline-shift": 10}
    # )

    # vcv = tree.pcm.get_vcv_matrix_from_tree()
    # rng = np.random.default_rng(123)
    # print(rng.multivariate_normal([0] * tree.nnodes, 1. * vcv))
